@@ -407,9 +407,8 @@ class Command(object):
 
     def reset_bmc(self):
         """Do a cold reset in BMC"""
-
-        response = self.raw_command(netfn=6, command=2)
-        if 'error' in response:
+        response = self.raw_command(netfn=6, command=2, retry=False)
+        if response and 'error' in response:
             raise exc.IpmiException(response['error'])
 
     def set_bootdev(self,
@@ -870,6 +869,48 @@ class Command(object):
         self.oem_init()
         self._oem.clear_system_configuration()
 
+    def set_net6_configuration(self, static_addresses=None, static_gateway=None, channel=None):
+        if static_addresses is None and static_gateway is None:
+            return
+        if channel is None:
+            channel = self.get_network_channel()
+        if static_addresses is not None:
+            i = 0
+            for va in static_addresses:
+                if '/' in va:
+                    va, plen = va.split('/', 1)
+                else:
+                    plen = '64'
+                plen = int(plen)
+                vab = bytearray(socket.inet_pton(socket.AF_INET6, va))
+                cmddata = bytearray([channel, 56, 0, 0x80]) + vab + bytearray([plen, 0])
+                self.xraw_command(netfn=0xc, command=1, data=cmddata)
+        if static_gateway is not None:
+            gwb = bytearray(socket.inet_pton(socket.AF_INET6, static_gateway))
+            cmddata = bytearray([channel, 65]) + gwb
+            self.xraw_command(netfn=0xc, command=1, data=cmddata)
+    
+    def get_net6_configuration(self, channel=None):
+        if channel is None:
+            channel = self.get_network_channel()
+        retdata = {}
+        ip6a = self.xraw_command(netfn=0xc, command=2, data=(channel, 56, 0, 0))
+        ip6d = bytearray(ip6a['data'])
+        if ip6d[0] != 0x11:
+            raise Exception('Unsupported reply')
+        if ip6d[2] & 0x80 == 0x80:
+            ip6b = ip6d[3:19]
+            ip6addr = socket.inet_ntop(socket.AF_INET6, ip6b)
+            plen = ip6d[19]
+            retdata['static_addrs'] = ['{}/{}'.format(ip6addr, plen)]
+        ip6g = self.xraw_command(netfn=0xc, command=2, data=(channel, 65, 0, 0))
+        ip6gd = bytearray(ip6g['data'])
+        if ip6gd[0] != 0x11:
+            raise Exception('Unsupported reply')
+        gwa = socket.inet_ntop(socket.AF_INET6, ip6gd[1:17])
+        retdata['static_gateway'] = gwa
+        return retdata
+
     def set_net_configuration(self, ipv4_address=None, ipv4_configuration=None,
                               ipv4_gateway=None, channel=None):
         """Set network configuration data.
@@ -884,6 +925,9 @@ class Command(object):
         :param ipv4_gateway: IP address of gateway to use.
         :param channel:  LAN channel to configure, defaults to autodetect
         """
+        if (ipv4_address is None and ipv4_configuration is None
+                and ipv4_gateway is None):
+            return
         if channel is None:
             channel = self.get_network_channel()
         if ipv4_configuration is not None:
@@ -1699,6 +1743,7 @@ class Command(object):
             * administrator
             * proprietary
             * no_access
+            * custom.<name>
         """
         self.oem_init()
         if hasattr(self._oem, 'oem_user_access'):
@@ -1727,6 +1772,8 @@ class Command(object):
         self.oem_init()
         self._oem.set_user_access(
             uid, channel, callback, link_auth, ipmi_msg, privilege_level)
+        if privilege_level.startswith('custom.'):
+            return True  # unable to proceed with standard support
         data = [b, uid & 0b00111111,
                 privilege_levels[privilege_level] & 0b00001111, 0]
         response = self.raw_command(netfn=0x06, command=0x43, data=data)
@@ -1773,16 +1820,21 @@ class Command(object):
         r['access']['callback'] = (data[3] & 0b01000000) != 0
         r['access']['link_auth'] = (data[3] & 0b00100000) != 0
         r['access']['ipmi_msg'] = (data[3] & 0b00010000) != 0
-        privilege_levels = {
-            0: 'reserved',
-            1: 'callback',
-            2: 'user',
-            3: 'operator',
-            4: 'administrator',
-            5: 'proprietary',
-            0x0F: 'no_access'
-        }
-        r['access']['privilege_level'] = privilege_levels[data[3] & 0b00001111]
+        self.oem_init()
+        oempriv = self._oem.get_user_privilege_level(uid)
+        if oempriv:
+            r['access']['privilege_level'] = oempriv
+        else:
+            privilege_levels = {
+                0: 'reserved',
+                1: 'callback',
+                2: 'user',
+                3: 'operator',
+                4: 'administrator',
+                5: 'proprietary',
+                0x0F: 'no_access'
+            }
+            r['access']['privilege_level'] = privilege_levels[data[3] & 0b00001111]
         return r
 
     def set_user_name(self, uid, name):
@@ -2000,6 +2052,11 @@ class Command(object):
         """
         # TODO(jjohnson2): Provide OEM extensibility to cover user deletion
         self.oem_init()
+        if hasattr(self._oem, 'user_delete'):
+            try:
+                self._oem.user_delete(uid, channel)
+            except exc.BypassGenericBehavior:
+                return
         if hasattr(self._oem, 'user_delete_privilege_level'):
             privilege_level = self._oem.user_delete_privilege_level()
         else:
