@@ -18,10 +18,12 @@ This contains functions to manage the firmware configuration of Lenovo servers
 """
 
 import ast
+import base64
 import random
 import struct
 
 import six
+import time
 
 import pyghmi.exceptions as pygexc
 
@@ -43,6 +45,8 @@ WRITE_COMMAND = [0x03]
 CLOSE_COMMAND = [0x05]
 SIZE_COMMAND = [0x06]
 
+class Unsupported(Exception):
+    pass
 
 def fromstring(inputdata):
     if b'!entity' in inputdata.lower():
@@ -132,11 +136,15 @@ def _eval_conditional(expression, cfg, setting):
 
 
 class LenovoFirmwareConfig(object):
-    def __init__(self, ipmicmd):
+    def __init__(self, xc, useipmi=True):
         if not etree:
             raise Exception("python-lxml and python-eficompressor required "
                             "for this function")
-        self.connection = ipmicmd
+        if useipmi:
+            self.connection = xc.ipmicmd
+        else:
+            self.connection = None
+        self.xc = xc
 
     def imm_size(self, filename):
         data = bytearray()
@@ -264,15 +272,45 @@ class LenovoFirmwareConfig(object):
             cfgfilename = "config"
         options = {}
         data = None
-        for _ in range(0, 30):
-            filehandle = self.imm_open(cfgfilename)
-            size = self.imm_size(cfgfilename)
-            data = self.imm_read(filehandle, size)
-            self.imm_close(filehandle)
-            data = EfiCompressor.FrameworkDecompress(data, len(data))
-            if len(data) != 0:
-                break
-            self.connection.ipmi_session.pause(2)
+        if self.connection:
+            rsp = ({}, 200)
+        else:
+            rsp = self.xc.grab_redfish_response_with_status(
+                    '/redfish/v1/Managers/1')
+        if rsp[1] == 200:
+            if 'purley' not in rsp[0].get('Oem', {}).get('Lenovo', {}).get(
+                                    'release_name', 'unknown'):
+                rsp = self.xc.grab_redfish_response_with_status(
+                    '/redfish/v1/Systems/1/Actions/Oem/LenovoComputerSystem.DSReadFile',
+                    {'Action': 'DSReadFile', 'FileName': cfgfilename})
+            else:
+                rsp = (None, 500)
+        if rsp[1] == 200:
+            for _ in range(0, 30):
+                data = rsp[0]['Content']
+                data = base64.b64decode(data)
+                data = EfiCompressor.FrameworkDecompress(data, len(data))
+                if len(data) != 0:
+                    break
+                if self.connection:
+                    self.connection.ipmi_session.pause(2)
+                else:
+                    time.sleep(2)
+                rsp = self.xc.grab_redfish_response_with_status(
+                    '/redfish/v1/Systems/1/Actions/Oem/LenovoComputerSystem.DSReadFile',
+                    {'Action': 'DSReadFile', 'FileName': cfgfilename})
+        else:
+            if self.connection is None:
+                raise Unsupported('Not Supported')
+            for _ in range(0, 30):
+                filehandle = self.imm_open(cfgfilename)
+                size = self.imm_size(cfgfilename)
+                data = self.imm_read(filehandle, size)
+                self.imm_close(filehandle)
+                data = EfiCompressor.FrameworkDecompress(data, len(data))
+                if len(data) != 0:
+                    break
+                self.connection.ipmi_session.pause(2)
         if not data:
             raise Exception("BMC failed to return configuration information")
         xml = fromstring(data)
@@ -549,8 +587,23 @@ class LenovoFirmwareConfig(object):
 
         xml = etree.tostring(configurations)
         data = EfiCompressor.FrameworkCompress(xml, len(xml))
+        bdata = base64.b64encode(data).decode('utf8')
+        rsp = self.xc.grab_redfish_response_with_status(
+                '/redfish/v1/Managers/1')
+        if rsp[1] == 200:
+            if 'purley' not in rsp[0].get('Oem', {}).get('Lenovo', {}).get(
+                    'release_name', 'purley'):
+                rsp = self.xc.grab_redfish_response_with_status(
+                    '/redfish/v1/Systems/1/Actions/Oem/'
+                    'LenovoComputerSystem.DSWriteFile',
+                    {'Action': 'DSWriteFile', 'Resize': len(data),
+                     'FileName': 'asu_update.efi', 'Content': bdata})
+                if rsp[1] == 204:
+                    return True
+        if self.connection is None:
+            raise Unsupported('Not Supported')
         filehandle = self.imm_open("asu_update.efi", write=True,
-                                   size=len(data))
+                                size=len(data))
         self.imm_write(filehandle, len(data), data)
         stubread = len(data)
         if stubread > 8:
