@@ -383,10 +383,14 @@ class OEMHandler(generic.OEMHandler):
                 raise pygexc.InvalidParameterValue("Index must be 0 or 1")
             return self.tsmahandler.set_ntp_server(server, index)
         return None
+    
+    def user_delete(self, uid, channel):
+        if self.has_xcc:
+            return self.immhandler.user_delete(uid)
 
     def set_user_access(self, uid, channel, callback, link_auth, ipmi_msg,
                         privilege_level):
-        if self.is_fpc and self._fpc_variant == 2:
+        if self.is_fpc and self._fpc_variant != 6:
             self.smmhandler.set_user_priv(uid, privilege_level)
 
     @property
@@ -714,27 +718,28 @@ class OEMHandler(generic.OEMHandler):
                 fru['FRU Number'] = bextra[0]
                 fru['Revision'] = bextra[4]
                 macs = bextra[6]
-                macprefix = None
-                idx = 0
-                endidx = len(macs) - 5
-                macprefix = None
-                while idx < endidx:
-                    currmac = macs[idx:idx + 6]
-                    if not isinstance(currmac, bytearray):
-                        # invalid vpd format, abort attempts to extract
-                        # mac in this way
-                        break
-                    if currmac == b'\x00\x00\x00\x00\x00\x00':
-                        break
-                    # VPD may veer off, detect and break off
-                    if macprefix is None:
-                        macprefix = currmac[:3]
-                    elif currmac[:3] != macprefix:
-                        break
-                    ms = mac_format.format(*currmac)
-                    ifidx = idx / 6 + 1
-                    fru['MAC Address {0}'.format(ifidx)] = ms
-                    idx = idx + 6
+                if macs:
+                    macprefix = None
+                    idx = 0
+                    endidx = len(macs) - 5
+                    macprefix = None
+                    while idx < endidx:
+                        currmac = macs[idx:idx + 6]
+                        if not isinstance(currmac, bytearray):
+                            # invalid vpd format, abort attempts to extract
+                            # mac in this way
+                            break
+                        if currmac == b'\x00\x00\x00\x00\x00\x00':
+                            break
+                        # VPD may veer off, detect and break off
+                        if macprefix is None:
+                            macprefix = currmac[:3]
+                        elif currmac[:3] != macprefix:
+                            break
+                        ms = mac_format.format(*currmac)
+                        ifidx = idx / 6 + 1
+                        fru['MAC Address {0}'.format(ifidx)] = ms
+                        idx = idx + 6
             except (AttributeError, KeyError, IndexError):
                 pass
             if self.has_xcc and name and name.startswith('PSU '):
@@ -1198,7 +1203,7 @@ class OEMHandler(generic.OEMHandler):
         if self.has_xcc:
             return self.immhandler.get_bmc_configuration()
         if self.is_fpc:
-            return self.smmhandler.get_bmc_configuration()
+            return self.smmhandler.get_bmc_configuration(self._fpc_variant)
         if self.has_tsma:
             return self.tsmahandler.get_bmc_configuration()
         return super(OEMHandler, self).get_bmc_configuration()
@@ -1233,6 +1238,8 @@ class OEMHandler(generic.OEMHandler):
             return self.immhandler.clear_bmc_configuration()
         elif self.is_fpc:
             return self.smmhandler.clear_bmc_configuration()
+        elif self.has_tsma:
+            return self.tsmahandler.clear_bmc_configuration()
         return super(OEMHandler, self).clear_system_configuration()
 
     def clear_system_configuration(self):
@@ -1310,6 +1317,15 @@ class OEMHandler(generic.OEMHandler):
             return True
         return False
 
+    def get_user_privilege_level(self, uid):
+        if self.has_xcc:
+            return self.immhandler.get_user_privilege_level(uid)
+        return None
+
+    def set_user_access(self, uid, channel, callback, link_auth, ipmi_msg, privilege_level):
+        if self.has_xcc:
+            self.immhandler.set_user_access(uid, privilege_level)
+
     def process_zero_fru(self, zerofru):
         if (self.oemid['manufacturer_id'] == 19046
                 and self.oemid['product_id'] == 13616):
@@ -1342,6 +1358,8 @@ class OEMHandler(generic.OEMHandler):
                 else:
                     zerofru['UUID'] = util.decode_wireformat_uuid(
                         guiddata['data'])
+        if self.is_fpc:
+            self.smmhandler.augment_zerofru(zerofru, self._fpc_variant)
         return self.process_fru(zerofru)
 
     def get_ami_sensor_reading(self, sensorname):
