@@ -23,6 +23,7 @@ import six
 import pyghmi.constants as pygconst
 import pyghmi.exceptions as pygexc
 import pyghmi.ipmi.private.session as ipmisession
+import pyghmi.ipmi.private.util as util
 from pyghmi.ipmi import sdr
 import pyghmi.util.webclient as webclient
 
@@ -151,22 +152,23 @@ def fpc_get_nodeperm(ipmicmd, number, sz):
         rsp['data'] = b'\x00' + bytes(rsp['data'])
     elif len(rsp['data']) == 6:  # New FPC format
         rsp['data'] = bytes(rsp['data'][:2]) + bytes(rsp['data'][3:])
-    perminfo = bytearray(rsp['data'])[1]
+    permdata = bytearray(rsp['data'])
+    perminfo = permdata[1]
     if sz == 6:  # FPC
-        permfail = ('\x02', '\x03')
+        permfail = (2, 3)
     else:  # SMM
-        permfail = ('\x02',)
+        permfail = (2,)
     if perminfo & 0x20:
-        if rsp['data'][4] in permfail:
+        if permdata[4] in permfail:
             states.append('Insufficient Power')
             health = pygconst.Health.Failed
-        elif rsp['data'][3:5] != '\x00\x00':
+        elif rsp['data'][3:5] != '\x00\x00' and permdata[4] not in (0, 1):
             states.append('No Power Permission')
             health = pygconst.Health.Failed
     if perminfo & 0x40:
         states.append('Node Fault')
         health = pygconst.Health.Failed
-    if rsp['data'][3:5] == '\x00\x00':
+    if rsp['data'][3:5] == '\x00\x00' or permdata[4] == 0:
         states.append('Absent')
     return (health, states)
 
@@ -180,7 +182,7 @@ def get_psu_count(ipmicmd, variant):
     if variant == 0x26:
         mymsg = ipmicmd.xraw_command(netfn=0x32, command=0xa8)
         builddata = bytearray(mymsg['data'])
-        if builddata[13] == 3:
+        if builddata[13] in (3, 6):
             return 9
         else:
             return 6
@@ -401,6 +403,7 @@ class SMMClient(object):
         'password_login_failures': 'passwordFailAllowdNum',
         'password_min_length': 'passwordMinLength',
         'password_lockout_period': 'passwordLockoutTimePeriod',
+        'timezone': 'timeZone',
     }
 
     fanmodes = {
@@ -412,21 +415,26 @@ class SMMClient(object):
         5: 'Boosted',
     }
 
-    def get_bmc_configuration(self):
+    def get_bmc_configuration(self, variant):
         settings = {}
-        self.wc.request(
+        wc = self.wc
+        wc.request(
             'POST', '/data',
             ('get=passwordMinLength,passwordForceChange,passwordDurationDays,'
              'passwordExpireWarningDays,passwordChangeInterval,'
              'passwordReuseCheckNum,passwordFailAllowdNum,'
-             'passwordLockoutTimePeriod'))
-        rsp = self.wc.getresponse()
+             'passwordLockoutTimePeriod,timeZone'))
+        rsp = wc.getresponse()
         rspbody = rsp.read()
         accountinfo = fromstring(rspbody)
         for rule in self.rulemap:
             ruleinfo = accountinfo.find(self.rulemap[rule])
             if ruleinfo is not None:
-                settings[rule] = {'value': int(ruleinfo.text)}
+                try:
+                    val = int(ruleinfo.text)
+                except ValueError:
+                    val = ruleinfo.text
+                settings[rule] = {'value': val}
         dwc = self.ipmicmd.xraw_command(0x32, 0x94)
         dwc = bytearray(dwc['data'])
         if len(dwc) not in (3, 4) or dwc[0] == 1:
@@ -445,7 +453,10 @@ class SMMClient(object):
         powercfg = self.ipmicmd.xraw_command(0x32, 0xa2)
         powercfg = bytearray(powercfg['data'])
         if len(powercfg) == 5:
-            powercfg = powercfg[1:]
+            if variant and variant >> 5:
+                powercfg = powercfg[-2:]
+            else:
+                powercfg = powercfg[1:]
         val = powercfg[0]
         if val == 2:
             val = 'N+N'
@@ -483,6 +494,66 @@ class SMMClient(object):
                      'delivery is at the sustained capacity of the remaining '
                      'supplies.')
         }
+        try:
+            numbays = 12
+            try:
+                chassiscapinfo = self.ipmicmd.xraw_command(
+                    0x32, 0x9d, data=[13])
+            except pygexc.IpmiException as e:
+                if e.ipmicode == 201:  # this must be a 2U
+                    numbays = 4
+                    chassiscapinfo = self.ipmicmd.xraw_command(
+                        0x32, 0x9d, data=[5])
+                else:
+                    raise
+            retoffset = 0
+            if len(chassiscapinfo['data']) > 10:
+                retoffset = 1
+            chassiscapstate = self.ipmicmd.xraw_command(
+                0x32, 0xa0, data=[numbays + 1])
+            capstate = bool(chassiscapstate['data'][retoffset])
+            capmin, capmax, protcap, usercap, thermcap = struct.unpack(
+                '<HHHHH', chassiscapinfo['data'][retoffset:retoffset + 10])
+            settings['chassis_user_cap'] = {
+                'value': usercap,
+                'help': 'Specify a maximum wattage to consume, this specific '
+                        'system implements a range from {0} to {1}.'.format(
+                            capmin, capmax)
+            }
+            settings['chassis_user_cap_active'] = {
+                'value': 'Enable' if capstate else 'Disable',
+                'help': 'Specify whether the user capping setting should be '
+                        'used or not at the chassis level.',
+            }
+            for baynum in range(numbays):
+                baynum += 1
+                baycapinfo = self.ipmicmd.xraw_command(
+                    0x32, 0x9d, data=[baynum])
+                capmin, capmax, protcap, usercap, thermcap = struct.unpack(
+                    '<HHHHH', baycapinfo['data'][retoffset:retoffset + 10])
+                settings['bay{0}_user_cap'.format(baynum)] = {
+                    'value': usercap,
+                    'help': 'Specify a maximum wattage for the server in bay '
+                            '{0} to consume, this specific system implements '
+                            'a range from {1} to {2}'.format(
+                                baynum, capmin, capmax)
+                }
+                settings['bay{0}_protective_cap'.format(baynum)] = {
+                    'value': protcap,
+                    'help': 'Show the current protective cap for the system '
+                            'in bay {0}'.format(baynum)
+                }
+                baycapstate = self.ipmicmd.xraw_command(
+                    0x32, 0xa0, data=[baynum])
+                baycapstate = bool(baycapstate['data'][retoffset])
+                settings['bay{0}_user_cap_active'.format(baynum)] = {
+                    'value': 'Enable' if baycapstate else 'Disable',
+                    'help': 'Specify whether the user capping setting should '
+                            'be used or not for bay {0}'.format(baynum),
+                    'possible': ['Enable', 'Disable'],
+                }
+        except Exception:
+            pass
         try:
             dhcpsendname = self.ipmicmd.xraw_command(0xc, 0x2,
                                                      data=[1, 0xc5, 0, 0])
@@ -506,13 +577,61 @@ class SMMClient(object):
             }
         except Exception:
             pass
+        try:
+            chassisvpd = self.ipmicmd.xraw_command(0x32, 0xb0, data=(5, 0))
+            chassisvpd = bytearray(chassisvpd['data'][2:])
+            chassisvpd = bytes(chassisvpd).strip()
+            if not isinstance(chassisvpd, str):
+                chassisvpd = chassisvpd.decode('utf8')
+            settings['chassis_model'] = {
+                'value': chassisvpd,
+                'help': ('Configure the chassis model number')
+            }
+        except Exception:
+            pass
+        try:
+            chassisvpd = self.ipmicmd.xraw_command(0x32, 0xb0, data=(5, 1))
+            chassisvpd = bytearray(chassisvpd['data'][2:])
+            chassisvpd = bytes(chassisvpd).strip()
+            if not isinstance(chassisvpd, str):
+                chassisvpd = chassisvpd.decode('utf8')
+            settings['chassis_serial'] = {
+                'value': chassisvpd,
+                'help': ('Configure the chassis serial number')
+            }
+        except Exception:
+            pass
         return settings
 
+    def set_bay_cap(self, baynum, val):
+        payload = struct.pack('<BH', baynum, int(val))
+        self.ipmicmd.xraw_command(0x32, 0x9e, data=payload)
+
+    def augment_zerofru(self, zerofru, variant):
+        if variant & 0x20 != 0x20:
+            return
+        model = self.ipmicmd.xraw_command(
+            netfn=0x32, command=0xb0, data=[5, 11])['data'][2:]
+        zerofru['Product name'] = bytes(model).strip()
+        zerofru['Manufacturer'] = 'Lenovo'
+
+    def set_bay_cap_active(self, baynum, val):
+        currstate = self.ipmicmd.xraw_command(0x32, 0xa0, data=[baynum])
+        currstate = currstate['data']
+        if len(currstate) == 5:
+            currstate = currstate[1:]
+        savemode = currstate[3]
+        enable = val.lower().startswith('enable')
+        payload = [baynum, 1 if enable else 0, savemode]
+        self.ipmicmd.xraw_command(0x32, 0x9f, data=payload)
+
     def set_bmc_configuration(self, changeset, variant):
         rules = []
         powercfg = [None, None]
         sendhost = None
         sendvci = None
+        newserial = None
+        newmodel = None
         for key in changeset:
             if not key:
                 raise pygexc.InvalidParameterValue('Empty key is invalid')
@@ -544,6 +663,24 @@ class SMMClient(object):
                 sendvci = stringtoboolean(
                     changeset[key]['value'],
                     'dhcp_sends_vendor_class_identifier')
+            if fnmatch.fnmatch('chassis_serial', key.lower()):
+                newserial = changeset[key]['value']
+            if fnmatch.fnmatch('chassis_model', key.lower()):
+                newmodel = changeset[key]['value']
+            # Variant low 8 bits is the height in U of chassis, so double that
+            #  to get maxbays
+            numbays = (self.smm_variant & 0x0f) << 1
+            for bayn in range(1, numbays + 1):
+                if fnmatch.fnmatch('bay{0}_user_cap'.format(bayn),
+                                   key.lower()):
+                    self.set_bay_cap(bayn, changeset[key]['value'])
+                if fnmatch.fnmatch(
+                        'bay{0}_user_cap_active'.format(bayn), key.lower()):
+                    self.set_bay_cap_active(bayn, changeset[key]['value'])
+            if fnmatch.fnmatch('chassis_user_cap', key.lower()):
+                self.set_bay_cap(numbays + 1, changeset[key]['value'])
+            if fnmatch.fnmatch('chassis_user_cap_active', key.lower()):
+                self.set_bay_cap_active(numbays + 1, changeset[key]['value'])
             if fnmatch.fnmatch('fanspeed', key.lower()):
                 for mode in self.fanmodes:
                     byteval = mode
@@ -558,13 +695,16 @@ class SMMClient(object):
                             changeset[key]['value']))
         if rules:
             rules = 'set={0}'.format(','.join(rules))
-            self.wc.request('POST', '/data', rules)
-            self.wc.getresponse().read()
+            wc = self.wc
+            wc.request('POST', '/data', rules)
+            wc.getresponse().read()
         if powercfg != [None, None]:
             if variant != 6:
                 if None in powercfg:
                     currcfg = self.ipmicmd.xraw_command(0x32, 0xa2)
                     currcfg = bytearray(currcfg['data'])
+                    if variant and variant >> 5 and len(currcfg) == 5:
+                        currcfg = currcfg[-2:]
                     if powercfg[0] is None:
                         powercfg[0] = currcfg[0]
                     if powercfg[1] is None:
@@ -581,6 +721,14 @@ class SMMClient(object):
         if sendvci is not None:
             sendvci = 1 if sendvci else 0
             self.ipmicmd.xraw_command(0xc, 1, data=[1, 0xc6, sendvci])
+        if newserial:
+            newserial = newserial.ljust(10).encode('utf8')
+            cmdata = b'\x05\x01' + newserial
+            self.ipmicmd.xraw_command(0x32, 0xaf, data=cmdata)
+        if newmodel:
+            newmodel = newmodel.ljust(10).encode('utf8')
+            cmdata = b'\x05\x00' + newmodel
+            self.ipmicmd.xraw_command(0x32, 0xaf, data=cmdata)
 
     def set_user_priv(self, uid, priv):
         if priv.lower() == 'administrator':
@@ -588,18 +736,39 @@ class SMMClient(object):
             username = bytes(rsp['data']).rstrip(b'\x00')
             if not isinstance(username, str):
                 username = username.decode('utf8')
-            self.wc.request(
+            wc = self.wc
+            wc.request(
                 'POST', '/data', 'set=user({0},1,{1},511,,4,15,0)'.format(
                     uid, username))
-            rsp = self.wc.getresponse()
+            rsp = wc.getresponse()
             rsp.read()
 
     def reseat_bay(self, bay):
+        bay = int(bay)
+        if bay == -1:
+            self.ipmicmd.xraw_command(0x32, 0xf5)
+            return
+        if bay % 2 == 0:
+            # even node may be unable to reseat based on shared io
+            try:
+                rsp = self.ipmicmd.xraw_command(0x32, 0xc5, data=[1])
+            except Exception:
+                rsp = {'data': [1]}
+            rsp['data'] = bytearray(rsp['data'])
+            if rsp['data'][0] == 2:  # shared io
+                rsp = self.ipmicmd.xraw_command(0x32, 0xa7, data=[bay - 1])
+                rsp['data'] = bytearray(rsp['data'])
+                if rsp['data'][1] == 0x80:
+                    raise Exception('Unable to reseat bay {0} due to bay {1} '
+                                    'being on with shared IO'.format(
+                                        bay, bay - 1))
         self.ipmicmd.xraw_command(netfn=0x32, command=0xa4,
-                                  data=[int(bay), 2])
+                                  data=[bay, 2])
 
     def get_diagnostic_data(self, savefile, progress=None, autosuffix=False,
                             variant=None):
+        if variant == 6:
+            raise Exception('Service data not supported on FPC')
         rsp = self.ipmicmd.xraw_command(netfn=0x32, command=0xb1, data=[0])
         if bytearray(rsp['data'])[0] != 0:
             raise Exception("Service data generation already in progress")
@@ -618,21 +787,22 @@ class SMMClient(object):
             rsp = self.ipmicmd.xraw_command(netfn=0x34, command=0x12, data=[1])
             if progress:
                 progress({'phase': 'initializing', 'progress': initpct})
-        if self.wc is None:
+        wc = self.wc
+        if wc is None:
             raise Exception("Failed to connect to web api")
         if variant and variant >> 5:
-            url = '/preview/smm2-ffdc.tgz?ST1={0}'.format(self.st1)
+            url = '/preview/smm2-ffdc.tgz?ST1={0}'.format(wc.st1)
         else:
-            url = '/preview/smm-ffdc.tgz?ST1={0}'.format(self.st1)
+            url = '/preview/smm-ffdc.tgz?ST1={0}'.format(wc.st1)
         if autosuffix and not savefile.endswith('.tgz'):
             savefile += '-smm-ffdc.tgz'
-        fd = webclient.FileDownloader(self.wc, url, savefile)
+        fd = webclient.FileDownloader(wc, url, savefile)
         fd.start()
         while fd.isAlive():
             fd.join(1)
-            if progress and self.wc.get_download_progress():
+            if progress and wc.get_download_progress():
                 progress({'phase': 'download',
-                          'progress': 100 * self.wc.get_download_progress()})
+                          'progress': 100 * wc.get_download_progress()})
         if progress:
             progress({'phase': 'complete'})
         return savefile
@@ -654,6 +824,7 @@ class SMMClient(object):
     def get_webclient(self):
         cv = self.ipmicmd.certverify
         wc = webclient.SecureHTTPConnection(self.smm, 443, verifycallback=cv)
+        wc.vintage = util._monotonic_time()
         wc.connect()
         loginform = urlencode(
             {
@@ -674,13 +845,13 @@ class SMMClient(object):
             if 'renew' in data.text:
                 raise Exception("Account password has expired on remote "
                                 "device")
-        self.st1 = None
-        self.st2 = None
+        wc.st1 = None
+        wc.st2 = None
         for data in authdata.findall('st1'):
-            self.st1 = data.text
+            wc.st1 = data.text
         for data in authdata.findall('st2'):
-            self.st2 = data.text
-        if not self.st2:
+            wc.st2 = data.text
+        if not wc.st2:
             # This firmware puts tokens in the html file, parse that
             wc.request('GET', '/index.html')
             rsp = wc.getresponse()
@@ -691,12 +862,12 @@ class SMMClient(object):
                 indexhtml = indexhtml.decode('utf8')
             for line in indexhtml.split('\n'):
                 if '"ST1"' in line:
-                    self.st1 = line.split()[-1].replace(
+                    wc.st1 = line.split()[-1].replace(
                         '"', '').replace(',', '')
                 if '"ST2"' in line:
-                    self.st2 = line.split()[-1].replace(
+                    wc.st2 = line.split()[-1].replace(
                         '"', '').replace(',', '')
-        if not self.st2:
+        if not wc.st2:
             wc.request('GET', '/scripts/index.ajs')
             rsp = wc.getresponse()
             body = rsp.read()
@@ -706,19 +877,20 @@ class SMMClient(object):
                 body = body.decode('utf8')
             for line in body.split('\n'):
                 if '"ST1"' in line:
-                    self.st1 = line.split()[-1].replace(
+                    wc.st1 = line.split()[-1].replace(
                         '"', '').replace(',', '')
                 if '"ST2"' in line:
-                    self.st2 = line.split()[-1].replace(
+                    wc.st2 = line.split()[-1].replace(
                         '"', '').replace(',', '')
-        if not self.st2:
+        if not wc.st2:
             raise Exception('Unable to locate ST2 token')
-        wc.set_header('ST2', self.st2)
+        wc.set_header('ST2', wc.st2)
         return wc
 
     def set_hostname(self, hostname):
-        self.wc.request('POST', '/data', 'set=hostname:' + hostname)
-        rsp = self.wc.getresponse()
+        wc = self.wc
+        wc.request('POST', '/data', 'set=hostname:' + hostname)
+        rsp = wc.getresponse()
         if rsp.status != 200:
             raise Exception(rsp.read())
         rsp.read()
@@ -731,12 +903,13 @@ class SMMClient(object):
             return data.text
 
     def get_netinfo(self):
-        self.wc.request('POST', '/data', 'get=hostname')
-        rsp = self.wc.getresponse()
+        wc = self.wc
+        wc.request('POST', '/data', 'get=hostname')
+        rsp = wc.getresponse()
         data = rsp.read()
         if rsp.status == 400:
-            self.wc.request('POST', '/data?get=hostname', '')
-            rsp = self.wc.getresponse()
+            wc.request('POST', '/data?get=hostname', '')
+            rsp = wc.getresponse()
             data = rsp.read()
         if rsp.status != 200:
             raise Exception(data)
@@ -744,8 +917,9 @@ class SMMClient(object):
         return currinfo
 
     def set_domain(self, domain):
-        self.wc.request('POST', '/data', 'set=dnsDomain:' + domain)
-        rsp = self.wc.getresponse()
+        wc = self.wc
+        wc.request('POST', '/data', 'set=dnsDomain:' + domain)
+        rsp = wc.getresponse()
         if rsp.status != 200:
             raise Exception(rsp.read())
         rsp.read()
@@ -758,17 +932,19 @@ class SMMClient(object):
             return data.text
 
     def get_ntp_enabled(self, variant):
-        self.wc.request('POST', '/data', 'get=ntpOpMode')
-        rsp = self.wc.getresponse()
+        wc = self.wc
+        wc.request('POST', '/data', 'get=ntpOpMode')
+        rsp = wc.getresponse()
         info = fromstring(rsp.read())
         self.logout()
         for data in info.findall('ntpOpMode'):
             return data.text == '1'
 
     def set_ntp_enabled(self, enabled):
-        self.wc.request('POST', '/data', 'set=ntpOpMode:{0}'.format(
+        wc = self.wc
+        wc.request('POST', '/data', 'set=ntpOpMode:{0}'.format(
             1 if enabled else 0))
-        rsp = self.wc.getresponse()
+        rsp = wc.getresponse()
         result = rsp.read()
         if not isinstance(result, str):
             result = result.decode('utf8')
@@ -777,9 +953,10 @@ class SMMClient(object):
             raise Exception("Unrecognized result: " + result)
 
     def set_ntp_server(self, server, index):
-        self.wc.request('POST', '/data', 'set=ntpServer{0}:{1}'.format(
+        wc = self.wc
+        wc.request('POST', '/data', 'set=ntpServer{0}:{1}'.format(
             index + 1, server))
-        rsp = self.wc.getresponse()
+        rsp = wc.getresponse()
         result = rsp.read()
         if not isinstance(result, str):
             result = result.decode('utf8')
@@ -789,9 +966,10 @@ class SMMClient(object):
         return True
 
     def get_ntp_servers(self):
-        self.wc.request(
+        wc = self.wc
+        wc.request(
             'POST', '/data', 'get=ntpServer1,ntpServer2,ntpServer3')
-        rsp = self.wc.getresponse()
+        rsp = wc.getresponse()
         result = fromstring(rsp.read())
         srvs = []
         for data in result.findall('ntpServer1'):
@@ -805,7 +983,8 @@ class SMMClient(object):
 
     def update_firmware(self, filename, data=None, progress=None, bank=None):
         if progress is None:
-            progress = lambda x: True
+            def progress(x):
+                return True
         z = None
         if data and hasattr(data, 'read'):
             if zipfile.is_zipfile(data):
@@ -821,37 +1000,38 @@ class SMMClient(object):
                     data = z.open(filename)
                     break
         progress({'phase': 'upload', 'progress': 0.0})
-        self.wc.request('POST', '/data', 'set=fwType:10')  # SMM firmware
-        rsp = self.wc.getresponse()
+        wc = self.wc
+        wc.request('POST', '/data', 'set=fwType:10')  # SMM firmware
+        rsp = wc.getresponse()
         rsp.read()
-        url = '/fwupload/fwupload.esp?ST1={0}'.format(self.st1)
+        url = '/fwupload/fwupload.esp?ST1={0}'.format(wc.st1)
         fu = webclient.FileUploader(
-            self.wc, url, filename, data, formname='fileUpload',
+            wc, url, filename, data, formname='fileUpload',
             otherfields={'preConfig': 'on'})
         fu.start()
         while fu.isAlive():
             fu.join(3)
             if progress:
                 progress({'phase': 'upload',
-                          'progress': 100 * self.wc.get_upload_progress()})
+                          'progress': 100 * wc.get_upload_progress()})
         progress({'phase': 'validating', 'progress': 0.0})
         url = '/data'
-        self.wc.request('POST', url, 'get=fwVersion,spfwInfo')
-        rsp = self.wc.getresponse()
+        wc.request('POST', url, 'get=fwVersion,spfwInfo')
+        rsp = wc.getresponse()
         rsp.read()
         if rsp.status != 200:
             raise Exception('Error validating firmware')
         progress({'phase': 'apply', 'progress': 0.0})
-        self.wc.request('POST', '/data', 'set=securityrollback:1')
-        self.wc.getresponse().read()
-        self.wc.request('POST', '/data', 'set=fwUpdate:1')
-        rsp = self.wc.getresponse()
+        wc.request('POST', '/data', 'set=securityrollback:1')
+        wc.getresponse().read()
+        wc.request('POST', '/data', 'set=fwUpdate:1')
+        rsp = wc.getresponse()
         rsp.read()
         complete = False
         while not complete:
             ipmisession.Session.pause(3)
-            self.wc.request('POST', '/data', 'get=fwProgress,fwUpdate')
-            rsp = self.wc.getresponse()
+            wc.request('POST', '/data', 'get=fwProgress,fwUpdate')
+            rsp = wc.getresponse()
             progdata = rsp.read()
             if rsp.status != 200:
                 raise Exception('Error applying firmware')
@@ -866,6 +1046,8 @@ class SMMClient(object):
         return 'complete'
 
     def get_inventory_descriptions(self, ipmicmd, variant):
+        if variant >> 5 == 0:
+            return
         psucount = get_psu_count(ipmicmd, variant)
         for idx in range(psucount):
             yield 'PSU {}'.format(idx + 1)
@@ -908,13 +1090,15 @@ class SMMClient(object):
         return psui
 
     def logout(self):
-        self.wc.request('POST', '/data/logout', None)
-        rsp = self.wc.getresponse()
+        wc = self.wc
+        wc.request('POST', '/data/logout', None)
+        rsp = wc.getresponse()
         rsp.read()
         self._wc = None
 
     @property
     def wc(self):
-        if not self._wc or self._wc.broken:
+        if (not self._wc or self._wc.broken
+                or self._wc.vintage < util._monotonic_time() + 30):
             self._wc = self.get_webclient()
         return self._wc
