@@ -84,6 +84,12 @@ MAX_IDLE = 29
 # incorrect idle
 
 
+def unmatched(response, netfn, cmd):
+    if not response:
+        return True
+    return response['command'] != cmd or response['netfn'] != netfn
+
+
 def define_worker():
     class _IOWorker(threading.Thread):
         def join(self):
@@ -488,6 +494,9 @@ class Session(object):
                 else:
                     self.iterwaiters.append(onlogon)
             return
+        self.awaitingresponse = False
+        self.lastresponse = None
+        self.atomicop = threading.RLock()
         self.broken = False
         self.socket = None
         self.logged = 0
@@ -497,9 +506,11 @@ class Session(object):
         else:
             self.privlevel = 4
             self.autopriv = True
+        self.onlogpayload = None
+        self.onlogpayloadtype = None
         self.logoutexpiry = None
         self.autokeepalive = keepalive
-        self.maxtimeout = 3  # be aggressive about giving up on initial packet
+        self.maxtimeout = 2  # be aggressive about giving up on initial packet
         self.incommand = False
         self.nameonly = 16  # default to name only lookups in RAKP exchange
         self.servermode = False
@@ -554,6 +565,7 @@ class Session(object):
         # since our connection has failed retries
         # deregister our keepalive facility
         self.lastpayload = None
+        self.onlogpayload = None
         with util.protect(KEEPALIVE_SESSIONS):
             Session.keepalive_sessions.pop(self, None)
         with util.protect(WAITING_SESSIONS):
@@ -575,6 +587,13 @@ class Session(object):
     def onlogon(self, parameter):
         if 'error' in parameter:
             self._mark_broken(parameter['error'])
+        elif self.onlogpayload:
+            self._cmdwait()
+            payload = self.onlogpayload
+            payload_type = self.onlogpayloadtype
+            self.incommand = _monotonic_time() + self._getmaxtimeout()
+            self.onlogpayload = None
+            self.send_payload(payload, payload_type)
         while self.logonwaiters:
             waiter = self.logonwaiters.pop()
             waiter(parameter)
@@ -584,6 +603,7 @@ class Session(object):
         #                 I picked 'xCAT' minus 1 so that a hexdump of packet
         #                 would show xCAT
         self.localsid = 2017673555
+        self.remseqnumber = None
         self.confalgo = 0
         self.aeskey = None
         self.integrityalgo = 0
@@ -736,25 +756,36 @@ class Session(object):
         while incrementtime < self.maxtimeout:
             cumulativetime += incrementtime
             incrementtime += 1
-        return cumulativetime + 1
+        return (cumulativetime + 1) * (self.logontries + 1)
 
     def _cmdwait(self):
         while self._isincommand():
             _io_wait(self._isincommand(), self.sockaddr, self.evq)
 
-    def awaitresponse(self, retry):
-        while retry and self.lastresponse is None and self.logged:
-            timeout = self.expiration - _monotonic_time()
-            _io_wait(timeout, self.sockaddr)
-            while self.iterwaiters:
-                waiter = self.iterwaiters.pop()
-                waiter({'success': True})
-            self.process_pktqueue()
-            with util.protect(WAITING_SESSIONS):
-                if (self in self.waiting_sessions
-                        and self.expiration < _monotonic_time()):
-                    self.waiting_sessions.pop(self, None)
-                    self._timedout()
+    def awaitresponse(self, retry, netfn, command):
+        self.awaitingresponse = True
+        try:
+            alltimeout = _monotonic_time() + (self._getmaxtimeout() * 2)
+            while (retry and unmatched(self.lastresponse, netfn, command)
+                    and (self.logging or self.logged or self.onlogpayload)):
+                timeout = self.expiration - _monotonic_time()
+                _io_wait(timeout, self.sockaddr)
+                while self.iterwaiters:
+                    waiter = self.iterwaiters.pop()
+                    waiter({'success': True})
+                self.process_pktqueue()
+                if _monotonic_time() > alltimeout:
+                    self._mark_broken()
+                    raise exc.IpmiException('Session no longer connected')
+                with util.protect(WAITING_SESSIONS):
+                    if (self in self.waiting_sessions
+                            and self.expiration < _monotonic_time()):
+                        self.waiting_sessions.pop(self, None)
+                        if not self.lastpayload and not self.logging:
+                            return
+                        self._timedout()
+        finally:
+            self.awaitingresponse = False
 
     def raw_command(self,
                     netfn,
@@ -766,15 +797,19 @@ class Session(object):
                     timeout=None,
                     callback=None,
                     rslun=0):
+        while (self.logging and not self.logged
+                and _monotonic_time() < self.logoutexpiry):
+            self.pause(1)
         if not self.logged:
             if (self.logoutexpiry is not None
                     and _monotonic_time() > self.logoutexpiry):
                 self._mark_broken()
             raise exc.IpmiException('Session no longer connected')
-        self._cmdwait()
-        if not self.logged:
-            raise exc.IpmiException('Session no longer connected')
-        self.incommand = _monotonic_time() + self._getmaxtimeout()
+        with util.protect(self.atomicop):
+            self._cmdwait()
+            if not self.logged:
+                raise exc.IpmiException('Session no longer connected')
+            self.incommand = _monotonic_time() + self._getmaxtimeout()
         self.lastresponse = None
         if callback is None:
             self.ipmicallback = self._generic_callback
@@ -798,8 +833,9 @@ class Session(object):
         # behavior of only the constructor needing a callback.  From then on,
         # synchronous usage of the class acts in a greenthread style governed
         # by order of data on the network
-        self.awaitresponse(retry)
+        self.awaitresponse(retry, netfn + 1, command)
         lastresponse = self.lastresponse
+        self.lastresponse = None
         self.incommand = False
         while self.evq:
             self.evq.popleft().set()
@@ -963,7 +999,6 @@ class Session(object):
         if 'error' in response:
             self.onlogon(response)
             return
-        self.maxtimeout = 6  # we have a confirmed bmc, be more tenacious
         if response['code'] == 0xcc and self.ipmi15only is not None:
             # tried ipmi 2.0 against a 1.5 which should work, but some bmcs
             # thought 'reserved' meant 'must be zero'
@@ -1022,6 +1057,8 @@ class Session(object):
     def _req_priv_level(self):
         self.logged = 1
         self.logoutexpiry = None
+        self.maxtimeout = 4  # Switch quickly to the relogin recovery
+        self.logontries = 1  # have one relogin waiting to heal
         response = self.raw_command(netfn=0x6, command=0x3b,
                                     data=[self.privlevel])
         if response['code']:
@@ -1033,6 +1070,7 @@ class Session(object):
                                             data=[self.privlevel])
             if response['code']:
                 self.logged = 0
+                self.onlogpayload = None
                 self.logging = False
                 mysuffix = " while requesting privelege level %d for %s" % (
                     self.privlevel, self.userid)
@@ -1109,7 +1147,7 @@ class Session(object):
                                         data=[0x8e, self.privlevel])
 
     def login(self):
-        self.logontries = 5
+        self.logontries = 1
         self._initsession()
         self._get_channel_auth_cap()
 
@@ -1252,6 +1290,7 @@ class Session(object):
         def _keptalive(response):
             self._generic_callback(response)
             response = self.lastresponse
+            self.lastresponse = None
             self.incommand = False
             while self.evq:
                 self.evq.popleft().set()
@@ -1262,7 +1301,8 @@ class Session(object):
 
     def _keepalive(self):
         """Performs a keepalive to avoid idle disconnect"""
-
+        if self.awaitingresponse:
+            return
         try:
             keptalive = False
             if self._customkeepalives:
@@ -1331,7 +1371,7 @@ class Session(object):
                 iserver.pktqueue.append(qent)
                 iserver.process_pktqueue()
                 return
-            if (hasattr(self, 'remseqnumber')
+            if (hasattr(self, 'remseqnumber') and self.remseqnumber is not None
                     and remseqnumber < self.remseqnumber):
                 return -5  # remote sequence number is too low, reject it
             self.remseqnumber = remseqnumber
@@ -1409,7 +1449,7 @@ class Session(object):
             if sid != self.localsid:  # session id mismatch, drop it
                 return
             remseqnumber = struct.unpack("<I", bytes(data[10:14]))[0]
-            if (hasattr(self, 'remseqnumber')
+            if (hasattr(self, 'remseqnumber') and self.remseqnumber is not None
                     and (remseqnumber < self.remseqnumber)
                     and (self.remseqnumber != 0xffffffff)):
                 return
@@ -1453,7 +1493,7 @@ class Session(object):
 
     def _got_rmcp_response(self, data):
         # see RMCP+ open session response table
-        if not (self.sessioncontext and self.sessioncontext != "Established"):
+        if not (self.sessioncontext and self.sessioncontext != "ESTABLISHED"):
             return -9
             # ignore payload as we are not in a state valid it
         if data[0] != self.rmcptag:
@@ -1575,6 +1615,7 @@ class Session(object):
             payload=payload, payload_type=constants.payload_types['rakp3'])
 
     def _relog(self):
+        self.logging = True
         self._initsession()
         self.logontries -= 1
         return self._get_channel_auth_cap()
@@ -1715,13 +1756,23 @@ class Session(object):
         self.nowait = True
         self.timeout += 1
         if self.timeout > self.maxtimeout:
-            response = {'error': 'timeout', 'code': 0xffff}
-            self.ipmicallback(response)
-            self.nowait = False
-            self._mark_broken()
-            return
+            if not self.logontries:
+                response = {'error': 'timeout', 'code': 0xffff}
+                self.ipmicallback(response)
+                self.nowait = False
+                self._mark_broken()
+                return
+            else:
+                self.maxtimeout = 2
+                if self.sessioncontext == 'ESTABLISHED':
+                    self.onlogpayload = self.lastpayload
+                    self.onlogpayloadtype = self.last_payload_type
+                    self.maxtimeout = 6
+                self._relog()
+                return
         elif self.sessioncontext == 'FAILED':
             self.lastpayload = None
+            self.onlogpayload = None
             self.nowait = False
             return
         if self.sessioncontext == 'OPENSESSION':
@@ -1735,7 +1786,6 @@ class Session(object):
             # If we can't be sure which RAKP was dropped or if RAKP3/4 was just
             # delayed, the most reliable thing to do is rewind and start over
             # bmcs do not take kindly to receiving RAKP1 or RAKP3 twice
-            self.lastpayload = None
             self._relog()
         else:  # in IPMI case, the only recourse is to act as if the packet is
             # idempotent.  SOL has more sophisticated retry handling
@@ -1803,23 +1853,25 @@ class Session(object):
                 Session.waiting_sessions[self]['timeout'] = self.expiration
 
     def logout(self, sessionok=True):
-        if not self.logged:
-            return {'success': True}
+
         if self.cleaningup:
             self.nowait = True
-        if self.sol_handler:
-            self.raw_command(netfn=6, command=0x49, data=(1, 1, 0, 0, 0, 0),
-                             retry=sessionok)
-        self.raw_command(command=0x3c,
-                         netfn=6,
-                         data=struct.unpack("4B",
-                                            struct.pack("I", self.sessionid)),
-                         retry=False)
+        if self.logged:
+            if self.sol_handler:
+                self.raw_command(
+                    netfn=6, command=0x49, data=(1, 1, 0, 0, 0, 0),
+                    retry=sessionok)
+            self.raw_command(
+                command=0x3c, netfn=6,
+                data=struct.unpack(
+                    "4B", struct.pack("I", self.sessionid)),
+                retry=False)
         # stop trying for a keepalive,
         self.lastpayload = None
         with util.protect(KEEPALIVE_SESSIONS):
             Session.keepalive_sessions.pop(self, None)
         self.logged = 0
+        self.onlogpayload = None
         self.logging = False
         if self._customkeepalives:
             for ka in list(self._customkeepalives):
