@@ -41,10 +41,6 @@ import six
 import pyghmi.constants as const
 import pyghmi.exceptions as exc
 
-try:
-    import cPickle as pickle
-except ImportError:
-    import pickle
 
 TYPE_UNKNOWN = 0
 TYPE_SENSOR = 1
@@ -420,8 +416,6 @@ class SDREntry(object):
             # or hyphen, so go with space
             self.unit_mod = " "
         self.percent = ''
-        # two bytes of assertion event mask / lower threshold reading mask
-        self.assertion_reading_mask = entry[9:11]
         if entry[15] & 1 == 1:
             self.percent = '% '
         if self.sensor_type_number == 0xb:
@@ -516,43 +510,41 @@ class SDREntry(object):
         output['health'] = const.Health.Ok
         if discrete:
             for state in range(8):
-                if reading[2] & (0b1 << state) & \
-                        self.assertion_reading_mask[0]:
+                if reading[2] & (0b1 << state):
                     statedesc, health = self._decode_state(state)
                     output['health'] |= health
                     output['states'].append(statedesc)
                     output['state_ids'].append(self.assert_trap_value(state))
             if len(reading) > 3:
                 for state in range(7):
-                    if reading[3] & (0b1 << state) & \
-                            self.assertion_reading_mask[0]:
+                    if reading[3] & (0b1 << state):
                         statedesc, health = self._decode_state(state + 8)
                         output['health'] |= health
                         output['states'].append(statedesc)
                         output['state_ids'].append(
                             self.assert_trap_value(state + 8))
         else:
-            if reading[2] & 0b1 & self.assertion_reading_mask[0]:
+            if reading[2] & 0b1:
                 output['health'] |= const.Health.Warning
                 output['states'].append(lower + " non-critical threshold")
                 output['state_ids'].append(self.assert_trap_value(1))
-            if reading[2] & 0b10 & self.assertion_reading_mask[0]:
+            if reading[2] & 0b10:
                 output['health'] |= const.Health.Critical
                 output['states'].append(lower + " critical threshold")
                 output['state_ids'].append(self.assert_trap_value(2))
-            if reading[2] & 0b100 & self.assertion_reading_mask[0]:
+            if reading[2] & 0b100:
                 output['health'] |= const.Health.Failed
                 output['states'].append(lower + " non-recoverable threshold")
                 output['state_ids'].append(self.assert_trap_value(3))
-            if reading[2] & 0b1000 & self.assertion_reading_mask[0]:
+            if reading[2] & 0b1000:
                 output['health'] |= const.Health.Warning
                 output['states'].append(upper + " non-critical threshold")
                 output['state_ids'].append(self.assert_trap_value(4))
-            if reading[2] & 0b10000 & self.assertion_reading_mask[0]:
+            if reading[2] & 0b10000:
                 output['health'] |= const.Health.Critical
                 output['states'].append(upper + " critical threshold")
                 output['state_ids'].append(self.assert_trap_value(5))
-            if reading[2] & 0b100000 & self.assertion_reading_mask[0]:
+            if reading[2] & 0b100000:
                 output['health'] |= const.Health.Failed
                 output['states'].append(upper + " non-recoverable threshold")
                 output['state_ids'].append(self.assert_trap_value(6))
@@ -740,15 +732,17 @@ class SDR(object):
         cachefilename = None
         self.broken_sensor_ids = {}
         if self.cachedir:
-            cachefilename = 'sdrcache.{0}.{1}.{2}.{3}.{4}.{5}'.format(
+            cachefilename = 'sdrcache-2.{0}.{1}.{2}.{3}.{4}.{5}'.format(
                 self.mfg_id, self.prod_id, self.device_id, self.fw_major,
                 self.fw_minor, modtime)
             cachefilename = os.path.join(self.cachedir, cachefilename)
         if cachefilename and os.path.isfile(cachefilename):
             with open(cachefilename, 'rb') as cfile:
-                csdrs = pickle.load(cfile)
-                for sdrdata in csdrs:
-                    self.add_sdr(sdrdata)
+                csdrlen = cfile.read(2)
+                while csdrlen:
+                    csdrlen = struct.unpack('!H', csdrlen)[0]
+                    self.add_sdr(cfile.read(csdrlen))
+                    csdrlen = cfile.read(2)
                 for sid in self.broken_sensor_ids:
                     try:
                         del self.sensors[sid]
@@ -807,7 +801,7 @@ class SDR(object):
                     size = currlen - offset
             self.add_sdr(sdrdata)
             if sdrraw is not None:
-                sdrraw.append(sdrdata)
+                sdrraw.append(bytes(sdrdata))
             offset = 0
             if size != 0xff:
                 size = 5
@@ -828,7 +822,9 @@ class SDR(object):
             suffix = ''.join(
                 random.choice(string.ascii_lowercase) for _ in range(12))
             with open(cachefilename + '.' + suffix, 'wb') as cfile:
-                pickle.dump(sdrraw, cfile)
+                for csdr in sdrraw:
+                    cfile.write(struct.pack('!H', len(csdr)))
+                    cfile.write(csdr)
             os.rename(cachefilename + '.' + suffix, cachefilename)
 
     def get_sensor_numbers(self):
@@ -841,6 +837,8 @@ class SDR(object):
                         False, self.mfg_id, self.prod_id)
 
     def add_sdr(self, sdrbytes):
+        if not isinstance(sdrbytes[0], int):
+            sdrbytes = bytearray(sdrbytes)
         newent = self.make_sdr_entry(sdrbytes)
         if newent.sdrtype == TYPE_SENSOR:
             id = '{0}.{1}.{2}'.format(
