@@ -21,7 +21,6 @@ for redfish compliant endpoints
 import base64
 from datetime import datetime
 from datetime import timedelta
-from fnmatch import fnmatch
 import json
 import os
 import re
@@ -145,80 +144,6 @@ class SensorReading(object):
         self.unavailable = unavailable
 
 
-class AttrDependencyHandler(object):
-    def __init__(self, dependencies, currsettings, pendingsettings):
-        self.dependencymap = {}
-        for dep in dependencies.get('Dependencies', [{}]):
-            if 'Dependency' not in dep:
-                continue
-            if dep['Type'] != 'Map':
-                continue
-            if dep['DependencyFor'] in self.dependencymap:
-                self.dependencymap[
-                    dep['DependencyFor']].append(dep['Dependency'])
-            else:
-                self.dependencymap[
-                    dep['DependencyFor']] = [dep['Dependency']]
-        self.curr = currsettings
-        self.pend = pendingsettings
-        self.reg = dependencies['Attributes']
-
-    def get_overrides(self, setting):
-        overrides = {}
-        blameattrs = []
-        if setting not in self.dependencymap:
-            return {}, []
-        for depinfo in self.dependencymap[setting]:
-            lastoper = None
-            lastcond = None
-            for mapfrom in depinfo.get('MapFrom', []):
-                if lastcond is not None and not lastoper:
-                    break  # MapTerm required to make sense of this, give up
-                currattr = mapfrom['MapFromAttribute']
-                blameattrs.append(currattr)
-                currprop = mapfrom['MapFromProperty']
-                if currprop == 'CurrentValue':
-                    if currattr in self.pend:
-                        currval = self.pend[currattr]
-                    else:
-                        currval = self.curr[currattr]
-                else:
-                    currval = self.reg[currattr][currprop]
-                lastcond = self.process(currval, mapfrom, lastcond, lastoper)
-                lastoper = mapfrom.get('MapTerms', None)
-            if lastcond:
-                if setting not in overrides:
-                    overrides[setting] = {}
-                if depinfo['MapToAttribute'] not in overrides[setting]:
-                    overrides[depinfo['MapToAttribute']] = {}
-                overrides[depinfo['MapToAttribute']][
-                    depinfo['MapToProperty']] = depinfo['MapToValue']
-        return overrides, blameattrs
-
-    def process(self, currval, mapfrom, lastcond, lastoper):
-        newcond = None
-        mfc = mapfrom['MapFromCondition']
-        if mfc == 'EQU':
-            newcond = currval == mapfrom['MapFromValue']
-        if mfc == 'NEQ':
-            newcond = currval != mapfrom['MapFromValue']
-        if mfc == 'GEQ':
-            newcond = float(currval) >= float(mapfrom['MapFromValue'])
-        if mfc == 'GTR':
-            newcond = float(currval) > float(mapfrom['MapFromValue'])
-        if mfc == 'LEQ':
-            newcond = float(currval) <= float(mapfrom['MapFromValue'])
-        if mfc == 'LSS':
-            newcond = float(currval) < float(mapfrom['MapFromValue'])
-        if lastcond is not None:
-            if lastoper == 'AND':
-                return lastcond and newcond
-            elif lastoper == 'OR':
-                return lastcond or newcond
-            return None
-        return newcond
-
-
 class Command(object):
 
     def __init__(self, bmc, userid, password, verifycallback, sysurl=None,
@@ -239,6 +164,7 @@ class Command(object):
         self._oem = None
         self._gpool = pool
         self._bmcv4ip = None
+        self._bmcv6ip = None
         for addrinf in socket.getaddrinfo(bmc, 0, 0, socket.SOCK_STREAM):
             if addrinf[0] == socket.AF_INET:
                 self._bmcv4ip = socket.inet_pton(addrinf[0], addrinf[-1][0])
@@ -428,6 +354,8 @@ class Command(object):
         return True
 
     def set_user_access(self, uid, privilege_level='ReadOnly'):
+        if privilege_level.startswith('custom.'):
+            privilege_level = privilege_level.replace('custom.', '')
         accinfo = self._account_url_info_by_id(uid)
         if not accinfo:
             raise Exception("Unable to find indicated uid")
@@ -449,6 +377,8 @@ class Command(object):
         accinfo = self._account_url_info_by_id(uid)
         if not accinfo:
             raise Exception("Unable to find indicated uid")
+        if privilege_level.startswith('custom.'):
+            privilege_level = privilege_level.replace('custom.', '')
         for role in self._validroles:
             if role.lower() == privilege_level.lower():
                 privilege_level = role
@@ -463,13 +393,7 @@ class Command(object):
         return True
 
     def user_delete(self, uid):
-        # Redfish doesn't do so well with Deleting users either...
-        # Blanking the username seems to be the convention
-        # First, set a bogus password in case the implementation does honor
-        # blank user, at least render such an account harmless
-        self.set_user_password(uid, base64.b64encode(os.urandom(15)))
-        self.set_user_name(uid, '')
-        return True
+        self.oem.user_delete(uid)
 
     def set_user_name(self, uid, name):
         """Set user name
@@ -908,40 +832,6 @@ class Command(object):
             summary['badreadings'].append(unkinf)
         return summary
 
-    def _get_biosreg(self, url):
-        addon = {}
-        valtodisplay = {}
-        displaytoval = {}
-        reg = self._do_web_request(url)
-        reg = reg['RegistryEntries']
-        for attr in reg['Attributes']:
-            vals = attr.get('Value', [])
-            if vals:
-                valtodisplay[attr['AttributeName']] = {}
-                displaytoval[attr['AttributeName']] = {}
-                for val in vals:
-                    valtodisplay[
-                        attr['AttributeName']][val['ValueName']] = val[
-                            'ValueDisplayName']
-                    displaytoval[
-                        attr['AttributeName']][val['ValueDisplayName']] = val[
-                            'ValueName']
-            defaultval = attr.get('DefaultValue', None)
-            defaultval = valtodisplay.get(attr['AttributeName'], {}).get(
-                defaultval, defaultval)
-            if attr['Type'] == 'Integer' and defaultval:
-                defaultval = int(defaultval)
-            if attr['Type'] == 'Boolean':
-                vals = [{'ValueDisplayName': 'True'},
-                        {'ValueDisplayName': 'False'}]
-            addon[attr['AttributeName']] = {
-                'default': defaultval,
-                'help': attr.get('HelpText', None),
-                'sortid': attr.get('DisplayOrder', None),
-                'possible': [x['ValueDisplayName'] for x in vals],
-            }
-        return addon, valtodisplay, displaytoval, reg
-
     def get_bmc_configuration(self):
         """Get miscellaneous BMC configuration
 
@@ -966,6 +856,9 @@ class Command(object):
         # For now, this is a stub, no implementation for redfish currently
         return self.oem.set_bmc_configuration(changeset)
 
+    def set_system_configuration(self, changeset):
+        return self.oem.set_system_configuration(changeset, self)
+
     def clear_bmc_configuration(self):
         """Reset BMC to factory default
 
@@ -977,64 +870,7 @@ class Command(object):
             'Clear BMC configuration not supported in redfish yet')
 
     def get_system_configuration(self, hideadvanced=True):
-        return self._getsyscfg()[0]
-
-    def _getsyscfg(self):
-        biosinfo = self._do_web_request(self._biosurl, cache=False)
-        reginfo = ({}, {}, {}, {})
-        extrainfo = {}
-        valtodisplay = {}
-        self.attrdeps = {'Dependencies': [], 'Attributes': []}
-        if 'AttributeRegistry' in biosinfo:
-            overview = self._do_web_request('/redfish/v1/')
-            reglist = overview['Registries']['@odata.id']
-            reglist = self._do_web_request(reglist)
-            regurl = None
-            for cand in reglist.get('Members', []):
-                cand = cand.get('@odata.id', '')
-                candname = cand.split('/')[-1]
-                if candname == '':  # implementation uses trailing slash
-                    candname = cand.split('/')[-2]
-                if candname == biosinfo['AttributeRegistry']:
-                    regurl = cand
-                    break
-            if not regurl:
-                # Workaround a vendor bug where they link to a
-                # non-existant name
-                for cand in reglist.get('Members', []):
-                    cand = cand.get('@odata.id', '')
-                    candname = cand.split('/')[-1]
-                    candname = candname.split('.')[0]
-                    if candname == biosinfo[
-                            'AttributeRegistry'].split('.')[0]:
-                        regurl = cand
-                        break
-            if regurl:
-                reginfo = self._do_web_request(regurl)
-                for reg in reginfo.get('Location', []):
-                    if reg.get('Language', 'en').startswith('en'):
-                        reguri = reg['Uri']
-                        reginfo = self._get_biosreg(reguri)
-                        extrainfo, valtodisplay, _, self.attrdeps = reginfo
-        currsettings = {}
-        try:
-            pendingsettings = self._do_web_request(self._setbiosurl)
-        except exc.UnsupportedFunctionality:
-            pendingsettings = {}
-        pendingsettings = pendingsettings.get('Attributes', {})
-        for setting in biosinfo.get('Attributes', {}):
-            val = biosinfo['Attributes'][setting]
-            currval = val
-            if setting in pendingsettings:
-                val = pendingsettings[setting]
-            val = valtodisplay.get(setting, {}).get(val, val)
-            currval = valtodisplay.get(setting, {}).get(currval, currval)
-            val = {'value': val}
-            if currval != val['value']:
-                val['active'] = currval
-            val.update(**extrainfo.get(setting, {}))
-            currsettings[setting] = val
-        return currsettings, reginfo
+        return self.oem.get_system_configuration(hideadvanced, self)
 
     def clear_system_configuration(self):
         """Clear the BIOS/UEFI configuration
@@ -1063,82 +899,27 @@ class Command(object):
             parms = {'Action': 'Bios.ResetBios'}
         self._do_web_request(rb, parms)
 
-    def set_system_configuration(self, changeset):
-        while True:
-            try:
-                self._set_system_configuration(changeset)
-                return
-            except exc.RedfishError as re:
-                if ('etag' not in re.msgid.lower()
-                        and 'PreconditionFailed' not in re.msgid):
-                    raise
-
-    def _set_system_configuration(self, changeset):
-        currsettings, reginfo = self._getsyscfg()
-        rawsettings = self._do_web_request(self._biosurl, cache=False)
-        rawsettings = rawsettings.get('Attributes', {})
-        pendingsettings = self._do_web_request(self._setbiosurl)
-        etag = pendingsettings.get('@odata.etag', None)
-        pendingsettings = pendingsettings.get('Attributes', {})
-        dephandler = AttrDependencyHandler(self.attrdeps, rawsettings,
-                                           pendingsettings)
-        for change in list(changeset):
-            if change not in currsettings:
-                found = False
-                for attr in currsettings:
-                    if fnmatch(attr.lower(), change.lower()):
-                        found = True
-                        changeset[attr] = changeset[change]
-                    if fnmatch(attr.lower(),
-                               change.replace('.', '_').lower()):
-                        found = True
-                        changeset[attr] = changeset[change]
-                if found:
-                    del changeset[change]
-        for change in changeset:
-            changeval = changeset[change]
-            overrides, blameattrs = dephandler.get_overrides(change)
-            meta = {}
-            for attr in self.attrdeps['Attributes']:
-                if attr['AttributeName'] == change:
-                    meta = dict(attr)
-                    break
-            meta.update(**overrides.get(change, {}))
-            if meta.get('ReadOnly', False) or meta.get('GrayOut', False):
-                errstr = '{0} is read only'.format(change)
-                if blameattrs:
-                    errstr += (' due to one of the following settings: '
-                               '{0}'.format(','.join(sorted(blameattrs)))
-                               )
-                raise exc.InvalidParameterValue(errstr)
-            if (currsettings.get(change, {}).get('possible', [])
-                    and changeval not in currsettings[change]['possible']):
-                normval = changeval.lower()
-                normval = re.sub(r'\s+', ' ', normval)
-                if not normval.endswith('*'):
-                    normval += '*'
-                for cand in currsettings[change]['possible']:
-                    if fnmatch(cand.lower().replace(' ', ''),
-                               normval.replace(' ', '')):
-                        changeset[change] = cand
-                        break
+    def set_net6_configuration(self, static_addresses=None, static_gateway=None, name=None):
+        patch = {}
+        if static_addresses is not None:
+            sa = []
+            patch['IPv6StaticAddresses'] = sa
+            for addr in static_addresses:
+                if '/' in addr:
+                    addr, plen = addr.split('/', 1)
                 else:
-                    raise exc.InvalidParameterValue(
-                        '{0} is not a valid value for {1} ({2})'.format(
-                            changeval, change, ','.join(
-                                currsettings[change]['possible'])))
-            if changeset[change] in reginfo[2].get(change, {}):
-                changeset[change] = reginfo[2][change][changeset[change]]
-            for regentry in reginfo[3].get('Attributes', []):
-                if change in (regentry.get('AttributeName', ''),
-                              regentry.get('DisplayName', '')):
-                    if regentry.get('Type', None) == 'Integer':
-                        changeset[change] = int(changeset[change])
-                    if regentry.get('Type', None) == 'Boolean':
-                        changeset[change] = _to_boolean(changeset[change])
-        redfishsettings = {'Attributes': changeset}
-        self._do_web_request(self._setbiosurl, redfishsettings, 'PATCH',
-                             etag=etag)
+                    plen = '64'
+                sa.append({
+                    'PrefixLength': int(plen),
+                    'Address': addr
+                })
+        if static_gateway:
+            patch['IPv6StaticDefaultGateways'] = [{
+                'Address': static_gateway,
+            }]
+        if patch:
+            nicurl = self._get_bmc_nic_url(name)
+            self._do_web_request(nicurl, patch, 'PATCH')
 
     def set_net_configuration(self, ipv4_address=None, ipv4_configuration=None,
                               ipv4_gateway=None, name=None):
@@ -1146,6 +927,9 @@ class Command(object):
         ipinfo = {}
         dodhcp = None
         netmask = None
+        if (ipv4_address is None and ipv4_configuration is None
+                and ipv4_gateway is None):
+            return
         if ipv4_address:
             if '/' in ipv4_address:
                 ipv4_address, cidr = ipv4_address.split('/')
@@ -1159,7 +943,7 @@ class Command(object):
             patch['IPv4StaticAddresses'] = [ipinfo]
             ipinfo['Gateway'] = ipv4_gateway
             ipv4_configuration = 'static'
-        if ipv4_configuration.lower() == 'dhcp':
+        if ipv4_configuration and ipv4_configuration.lower() == 'dhcp':
             dodhcp = True
             patch['DHCPv4'] = {'DHCPEnabled': True}
         elif (ipv4_configuration == 'static'
@@ -1178,6 +962,21 @@ class Command(object):
                     ipinfo['AddressOrigin'] = 'Static'
                 self._do_web_request(nicurl, patch, 'PATCH')
 
+    def get_net6_configuration(self, name=None):
+        nicurl = self._get_bmc_nic_url(name)
+        netcfg = self._do_web_request(nicurl, cache=False)
+        retdata = {}
+        saddrs = netcfg.get('IPv6StaticAddresses', ())
+        retdata['static_addrs'] = []
+        for sa in saddrs:
+            ca = '{}/{}'.format(sa['Address'], sa['PrefixLength'])
+            retdata['static_addrs'].append(ca)
+        gws = netcfg.get('IPv6StaticDefaultGateways', None)
+        if gws:
+            for gw in gws:
+                retdata['static_gateway'] = gw['Address']
+        return retdata
+
     def get_net_configuration(self, name=None):
         nicurl = self._get_bmc_nic_url(name)
         netcfg = self._do_web_request(nicurl, cache=False)
@@ -1375,6 +1174,10 @@ class Command(object):
                         entries = self._do_web_request(entriesurl, cache=False)
                         newloginfo = self._do_web_request(lurl, cache=False)
             for log in entries.get('Members', []):
+                if ('Created' not in log and 'Message' not in log
+                        and 'Severity' not in log):
+                    # without any data, this log entry isn't actionable
+                    continue
                 record = {}
                 record['log_id'] = logid
                 parsedtime = parse_time(log.get('Created', ''))
@@ -1386,7 +1189,7 @@ class Command(object):
                     record['timestamp'] = log.get('Created', '')
                 record['message'] = log.get('Message', None)
                 record['severity'] = _healthmap.get(
-                    entries.get('Severity', 'Warning'), const.Health.Critical)
+                    log.get('Severity', 'Warning'), const.Health.Ok)
                 yield record
 
     def get_sensor_descriptions(self):
@@ -1465,6 +1268,16 @@ class Command(object):
         """
         return self.oem.apply_storage_configuration(cfgspec)
 
+    def check_storage_configuration(self, cfgspec=None):
+        """Evaluate a configuration for validity
+
+        This will check if configuration is currently available and, if given,
+        whether the specified cfgspec can be applied.
+        :param cfgspec: A pyghmi.storage.ConfigSpec describing desired oonfig
+        :return:
+        """
+        return self._oem.check_storage_configuration(cfgspec)
+
     def attach_remote_media(self, url, username=None, password=None):
         """Attach remote media by url
 
@@ -1481,9 +1294,12 @@ class Command(object):
         # identify the correct resource ahead of time.
         # As such it's OEM specific until the standard
         # provides a better way.
-        bmcinfo = self._do_web_request(self._bmcurl)
-        vmcoll = bmcinfo.get('VirtualMedia', {}).get('@odata.id', None)
-        vmurls = None
+        vmurls = []
+        vmcoll = self.sysinfo.get(
+            'VirtualMedia', {}).get('@odata.id', None)
+        if not vmcoll:
+            vmcoll = self.bmcinfo.get(
+                'VirtualMedia', {}).get('@odata.id', None)
         if vmcoll:
             vmlist = self._do_web_request(vmcoll)
             vmurls = [x['@odata.id'] for x in vmlist.get('Members', [])]
@@ -1513,12 +1329,14 @@ class Command(object):
             break
 
     def detach_remote_media(self):
-        bmcinfo = self._do_web_request(self._bmcurl)
-        vmcoll = bmcinfo.get('VirtualMedia', {}).get('@odata.id', None)
         try:
             self.oem.detach_remote_media()
         except exc.BypassGenericBehavior:
             return
+        vmcoll = self.sysinfo.get('VirtualMedia', {}).get('@odata.id', None)
+        if not vmcoll:
+            bmcinfo = self._do_web_request(self._bmcurl)
+            vmcoll = bmcinfo.get('VirtualMedia', {}).get('@odata.id', None)
         if vmcoll:
             vmlist = self._do_web_request(vmcoll)
             vmurls = [x['@odata.id'] for x in vmlist.get('Members', [])]
