@@ -14,6 +14,7 @@
 
 
 import errno
+import fnmatch
 import json
 import math
 import os
@@ -26,6 +27,7 @@ import six
 
 import pyghmi.exceptions as pygexc
 import pyghmi.ipmi.private.util as util
+import pyghmi.ipmi.oem.lenovo.config as config
 import pyghmi.media as media
 import pyghmi.redfish.oem.generic as generic
 import pyghmi.storage as storage
@@ -43,6 +45,22 @@ funtypes = {
     12: 'Fabric Controller',
 }
 
+tls_ver = {
+    1: 'TLS1.1',
+    2: 'TLS1.2',
+    3: 'TLS1.3'
+}
+
+sec_mode = {
+    0: ('Compatible',
+        'Maximum compatibility of cipher suites, without NIST compliance'),
+    1: ('NIST',
+        'Maximum compatibility within the constraints of being NIST '
+        'compliant with perfect forward secrecy'),
+    2: ('High', 'Support only modern and strong ciphers, NIST '
+                'compliant with perfect forward security'),
+}
+
 
 def naturalize_string(key):
     """Analyzes string in a human way to enable natural sort
@@ -92,8 +110,137 @@ class OEMHandler(generic.OEMHandler):
         super(OEMHandler, self).__init__(sysinfo, sysurl, webclient, cache,
                                          gpool)
         self._wc = None
+        self.weblogging = False
         self.updating = False
         self.datacache = {}
+        self.fwc = None
+        self.fwo = None
+
+    def get_system_configuration(self, hideadvanced=True, fishclient=None,
+                                 fetchimm=False):
+        if not self.fwc:
+            self.fwc = config.LenovoFirmwareConfig(self, useipmi=False)
+        try:
+            self.fwo = self.fwc.get_fw_options(fetchimm=fetchimm)
+        except config.Unsupported:
+            return super(OEMHandler, self).get_system_configuration(
+                hideadvanced, fishclient)
+        except Exception:
+            raise Exception('%s failed to retrieve UEFI configuration'
+                            % self.bmcname)
+        self.fwovintage = util._monotonic_time()
+        retcfg = {}
+        for opt in self.fwo:
+            if 'MegaRAIDConfigurationTool' in opt:
+                # Suppress the Avago configuration to be consistent with
+                # other tools.
+                continue
+            if (hideadvanced and self.fwo[opt]['lenovo_protect']
+                    or self.fwo[opt]['hidden']):
+                # Do not enumerate hidden settings
+                continue
+            retcfg[opt] = {}
+            retcfg[opt]['value'] = self.fwo[opt]['current']
+            retcfg[opt]['default'] = self.fwo[opt]['default']
+            retcfg[opt]['help'] = self.fwo[opt]['help']
+            retcfg[opt]['possible'] = self.fwo[opt]['possible']
+            retcfg[opt]['sortid'] = self.fwo[opt]['sortid']
+        return retcfg
+
+    def set_system_configuration(self, changeset, fishclient):
+        if not self.fwc:
+            self.fwc = config.LenovoFirmwareConfig(self, useipmi=False)
+        fetchimm = False
+        if not self.fwo or util._monotonic_time() - self.fwovintage > 30:
+            try:
+                self.fwo = self.fwc.get_fw_options(fetchimm=fetchimm)
+            except config.Unsupported:
+                return super(OEMHandler, self).set_system_configuration(
+                    changeset, fishclient)
+            self.fwovintage = util._monotonic_time()
+        for key in list(changeset):
+            if key not in self.fwo:
+                found = False
+                for rkey in self.fwo:
+                    if fnmatch.fnmatch(rkey.lower(), key.lower()):
+                        changeset[rkey] = changeset[key]
+                        found = True
+                    elif self.fwo[rkey].get('alias', None) != rkey:
+                        calias = self.fwo[rkey]['alias']
+                        if fnmatch.fnmatch(calias.lower(), key.lower()):
+                            changeset[rkey] = changeset[key]
+                            found = True
+                if not found and not fetchimm:
+                    fetchimm = True
+                    self.fwo = self.fwc.get_fw_options(fetchimm=fetchimm)
+                    if key in self.fwo:
+                        continue
+                    else:
+                        found = False
+                        for rkey in self.fwo:
+                            if fnmatch.fnmatch(rkey.lower(), key.lower()):
+                                changeset[rkey] = changeset[key]
+                                found = True
+                            elif self.fwo[rkey].get('alias', None) != rkey:
+                                calias = self.fwo[rkey]['alias']
+                                if fnmatch.fnmatch(
+                                        calias.lower(), key.lower()):
+                                    changeset[rkey] = changeset[key]
+                                    found = True
+                if found:
+                    del changeset[key]
+                else:
+                    raise pygexc.InvalidParameterValue(
+                        '{0} not a known setting'.format(key))
+        self.merge_changeset(changeset)
+        if changeset:
+            try:
+                self.fwc.set_fw_options(self.fwo)
+            finally:
+                self.fwo = None
+                self.fwovintage = 0
+
+    def merge_changeset(self, changeset):
+        for key in changeset:
+            if isinstance(changeset[key], six.string_types):
+                changeset[key] = {'value': changeset[key]}
+            newvalue = changeset[key]['value']
+            if self.fwo[key]['is_list'] and not isinstance(newvalue, list):
+                if '=' in newvalue:
+                    # ASU set a precedent of = delimited settings
+                    # for now, honor that delimiter as well
+                    newvalues = newvalue.split('=')
+                else:
+                    newvalues = newvalue.split(',')
+            else:
+                newvalues = [newvalue]
+            newnewvalues = []
+            for newvalue in newvalues:
+                newv = re.sub(r'\s+', ' ', newvalue)
+                if (self.fwo[key]['possible']
+                        and newvalue not in self.fwo[key]['possible']):
+                    candlist = []
+                    for candidate in self.fwo[key]['possible']:
+                        candid = re.sub(r'\s+', ' ', candidate)
+                        if newv.lower().startswith(candid.lower()):
+                            newvalue = candidate
+                            break
+                        if candid.lower().startswith(newv.lower()):
+                            candlist.append(candidate)
+                    else:
+                        if len(candlist) == 1:
+                            newvalue = candlist[0]
+                        else:
+                            raise pygexc.InvalidParameterValue(
+                                '{0} is not a valid value for {1} '
+                                '({2})'.format(
+                                    newvalue, key,
+                                    ','.join(self.fwo[key]['possible'])))
+                newnewvalues.append(newvalue)
+            if len(newnewvalues) == 1:
+                self.fwo[key]['new_value'] = newnewvalues[0]
+            else:
+                self.fwo[key]['new_value'] = newnewvalues
 
     def reseat_bay(self, bay):
         if bay != -1:
@@ -137,6 +284,50 @@ class OEMHandler(generic.OEMHandler):
             settings['presence_assert'] = {
                 'value': 'Enable' if asserted else 'Disable'
             }
+        secparms = self.wc.grab_json_response('/api/providers/imm_tls_mode')
+        if secparms and secparms.get('return', 1) == 0:
+            tlsmode = secparms.get('tls_ver', -1)
+            if tlsmode in tls_ver:
+                settings['minimum_tls_version'] = {
+                    'value': tls_ver[tlsmode],
+                    'help': 'The minimum TLS level allowed by the XCC when '
+                            'communicating',
+                    'possible': [tls_ver[x] for x in tls_ver],
+                }
+            secmode = secparms.get('sec_mode', -1)
+            if secmode in sec_mode:
+                settings['cryptography_mode'] = {
+                    'value': sec_mode[secmode][0],
+                    'help': 'Select cryptography mode, compatible allows all '
+                            'supported ciphers, NIST restricts to NIST '
+                            'compliant ciphers, and High further restricts '
+                            'to a modern subset of NIST compliance',
+                    'possible': [sec_mode[x][0] for x in sec_mode],
+                }
+        usbparms = self.wc.grab_json_response('/api/dataset/imm_usb')
+        if usbparms:
+            usbparms = usbparms.get('items', [{}])[0]
+            if usbparms['usb_eth_over_usb_enabled'] == 1:
+                usbeth = 'Enable'
+            else:
+                usbeth = 'Disable'
+            settings['usb_ethernet'] = {
+                'value': usbeth
+            }
+            if usbparms['usb_eth_to_eth_enabled'] == 1:
+                fwd = 'Enable'
+            else:
+                fwd = 'Disable'
+            settings['usb_ethernet_port_forwarding'] = {
+                'value': fwd
+            }
+            mappings = []
+            for mapping in usbparms['usb_mapped_ports']:
+                src = mapping['ext_port']
+                dst = mapping['eth_port']
+                if src != 0 and dst != 0:
+                    mappings.append('{0}:{1}'.format(src, dst))
+            settings['usb_forwarded_ports'] = {'value': ','.join(mappings)}
         return settings
 
     rulemap = {
@@ -151,6 +342,8 @@ class OEMHandler(generic.OEMHandler):
 
     def set_bmc_configuration(self, changeset):
         ruleset = {}
+        usbsettings = {}
+        secparms = {}
         for key in changeset:
             if isinstance(changeset[key], six.string_types):
                 changeset[key] = {'value': changeset[key]}
@@ -173,12 +366,117 @@ class OEMHandler(generic.OEMHandler):
                     raise pygexc.InvalidParameterValue(
                         '"{0}" is not a recognized value for {1}'.format(
                             currval, key))
+            elif key.lower() in (
+                    'usb_ethernet', 'usb_ethernet_port_forwarding',
+                    'usb_forwarded_ports'):
+                usbsettings[key] = changeset[key]['value']
+            elif key.lower() in ('cryptography_mode', 'minimum_tls_version'):
+                secparms[key] = changeset[key]['value']
             else:
                 raise pygexc.InvalidParameterValue(
                     '{0} not a known setting'.format(key))
+        if usbsettings:
+            self.apply_usb_configuration(usbsettings)
+        if secparms:
+            self.apply_sec_configuration(secparms)
         if ruleset:
             self.wc.grab_json_response('/api/dataset', ruleset)
 
+    def apply_sec_configuration(self, secparms):
+        secmode = None
+        tlsver = None
+        if 'cryptography_mode' in secparms:
+            secmode = secparms['cryptography_mode'].lower()
+            for sm in sec_mode:
+                if sec_mode[sm][0].lower().startswith(secmode):
+                    secmode = sm
+                    break
+            else:
+                raise pygexc.InvalidParameterValue(
+                    '"{0}" is not a recognized cryptography mode'.format(
+                        secparms['cryptography_mode']
+                    )
+                )
+        if 'minimum_tls_version' in secparms:
+            tlsver = secparms['minimum_tls_version'].lower()
+            for tv in tls_ver:
+                if tls_ver[tv].lower() == tlsver:
+                    tlsver = tv
+                    break
+            else:
+                raise pygexc.InvalidParameterValue(
+                    '"{0}" is not a recognized TLS version'.format(
+                        secparms['minimum_tls_version'])
+                )
+        if len(secparms) < 2:
+            currsecparms = self.wc.grab_json_response(
+                '/api/providers/imm_tls_mode')
+            del currsecparms['return']
+        else:
+            currsecparms = {}
+        if secmode is not None:
+            currsecparms['sec_mode'] = secmode
+        if tlsver is not None:
+            currsecparms['tls_ver'] = tlsver
+        self.wc.grab_json_response('/api/providers/imm_tls_mode',
+                                   currsecparms)
+
+    def apply_usb_configuration(self, usbsettings):
+        def numify(val):
+            if 'enabled'.startswith(val.lower()):
+                return '1'
+            if 'disabled'.startswith(val.lower()):
+                return '0'
+            raise Exception('Usupported value')
+        usbparms = self.wc.grab_json_response('/api/dataset/imm_usb')
+        usbparms = usbparms.get('items', [{}])[0]
+        addrmode = '{0}'.format(usbparms['lan_over_usb_addr_mode'])
+        ethena = '{0}'.format(usbparms['usb_eth_over_usb_enabled'])
+        fwdena = '{0}'.format(usbparms['usb_eth_to_eth_enabled'])
+        newena = usbsettings.get('usb_ethernet', None)
+        newfwd = usbsettings.get('usb_ethernet_port_forwarding', None)
+        newsettings = {
+            'USB_LANOverUSBAddrMode': addrmode,
+            'USB_EthOverUsbEna': ethena,
+            'USB_PortForwardEna': fwdena,
+            'USB_IPChangeEna': '0',
+        }
+        needsettings = False
+        if newena is not None:
+            needsettings = True
+            newsettings['USB_EthOverUsbEna'] = numify(newena)
+        if newfwd is not None:
+            needsettings = True
+            newsettings['USB_PortForwardEna'] = numify(newfwd)
+        if needsettings:
+            self.wc.grab_json_response('/api/dataset', newsettings)
+        if 'usb_forwarded_ports' in usbsettings:
+            oldfwds = {}
+            usedids = set([])
+            newfwds = usbsettings['usb_forwarded_ports'].split(',')
+            for mapping in usbparms['usb_mapped_ports']:
+                rule = '{0}:{1}'.format(
+                    mapping['ext_port'], mapping['eth_port'])
+                if rule not in newfwds:
+                    self.wc.grab_json_response(
+                        '/api/function', {
+                            'USB_RemoveMapping': '{0}'.format(mapping['id'])})
+                else:
+                    usedids.add(mapping['id'])
+                    oldfwds[rule] = mapping['id']
+            for mapping in usbsettings['usb_forwarded_ports'].split(','):
+                if mapping not in oldfwds:
+                    newid = 1
+                    while newid in usedids:
+                        newid += 1
+                    if newid > 11:
+                        raise Exception('Too Many Port Forwards')
+                    usedids.add(newid)
+                    newmapping = '{0},{1}'.format(
+                        newid, mapping.replace(':', ','))
+                    self.wc.grab_json_response(
+                        '/api/function', {'USB_AddMapping': newmapping})
+
     def get_description(self):
         description = self._do_web_request('/DeviceDescription.json')
         if description:
@@ -195,8 +493,11 @@ class OEMHandler(generic.OEMHandler):
         return {'height': u_height, 'slot': slot}
 
     def _get_agentless_firmware(self, components):
+        skipkeys = set([])
         adata = self.wc.grab_json_response(
             '/api/dataset/imm_adapters?params=pci_GetAdapters')
+        fdata = self.wc.grab_json_response(
+            '/api/function/adapter_update?params=pci_GetAdapterListAndFW')
         anames = set()
         for adata in adata.get('items', []):
             baseaname = adata['adapterName']
@@ -229,6 +530,24 @@ class OEMHandler(generic.OEMHandler):
                         except ValueError:
                             pass
                     yield '{0} {1}'.format(aname, fname), bdata
+            for fwi in fdata.get('items', []):
+                if fwi.get('key', -1) == adata.get('key', -2):
+                    skipkeys.add(fwi['key'])
+                    if fwi.get('fw_status', 0) == 2:
+                        bdata = {}
+                        if 'fw_version_pend' in fwi:
+                            bdata['version'] = fwi['fw_version_pend']
+                        yield '{0} Pending Update'.format(aname), bdata
+        for fwi in fdata.get('items', []):
+            if fwi.get('key', -1) > 0 and fwi['key'] not in skipkeys:
+                bdata = {}
+                bdata['version'] = fwi['fw_version']
+                yield fwi['adapterName'], bdata
+                if fwi.get('fw_status', 0) == 2:
+                    bdata = {}
+                    if 'fw_version_pend' in fwi:
+                        bdata['version'] = fwi['fw_version_pend']
+                    yield '{0} Pending Update'.format(fwi['adapterName']), bdata
 
     def _get_disk_firmware_single(self, diskent, prefix=''):
         bdata = {}
@@ -270,6 +589,8 @@ class OEMHandler(generic.OEMHandler):
                     yield ('LXPM Windows Driver Bundle', firminfo)
                 elif firm['type'] == 9:
                     yield ('LXPM Linux Driver Bundle', firminfo)
+                elif firm['type'] == 10:
+                    yield ('LXUM', firminfo)
         for adpinfo in self._get_agentless_firmware(components):
             yield adpinfo
         for adpinfo in self._get_disk_firmware(components):
@@ -404,6 +725,19 @@ class OEMHandler(generic.OEMHandler):
             themap[raidname] = mapdata
         return themap
 
+    def check_storage_configuration(self, cfgspec=None):
+        rsp = self.wc.grab_json_response(
+            '/api/function/raid_conf?params=raidlink_GetStatus')
+        if rsp['items'][0]['status'] not in (2, 3):
+            raise pygexc.TemporaryError('Storage configuration unavailable in '
+                                        'current state (try boot to setup or '
+                                        'an OS)')
+        if not cfgspec:
+            return True
+        for pool in cfgspec.arrays:
+            self._parse_storage_cfgspec(pool)
+        return True
+
     def _wait_storage_async(self):
         rsp = {'items': [{'status': 0}]}
         while rsp['items'][0]['status'] == 0:
@@ -477,9 +811,23 @@ class OEMHandler(generic.OEMHandler):
     def _create_array(self, pool):
         params = self._parse_array_spec(pool)
         cid = params['controller'].split(',')[0]
+        cslotno = params['controller'].split(',')[1]
         url = '/api/function/raid_conf?params=raidlink_GetDefaultVolProp'
         args = (url, cid, 0, params['drives'])
         props = self.wc.grab_json_response(','.join([str(x) for x in args]))
+        usesctrlslot = False
+        if not props:  # newer firmware requires raidlevel too
+            args = (url, cid, params['raidlevel'], 0, params['drives'])
+            props = self.wc.grab_json_response(','.join([str(x) for x in args]))
+        elif 'return' in props and props['return'] == 22:
+            # Jan 2023 XCC FW - without controller slot number
+            args = (url, cid, params['raidlevel'], 0, params['drives'])
+            props = self.wc.grab_json_response(','.join([str(x) for x in args]))
+            if 'return' in props and props['return'] == 22:
+                usesctrlslot = True
+                # Jan 2023 XCC FW - with controller slot number
+                args = (url, cid, params['raidlevel'], 0, params['drives'], cslotno)
+                props = self.wc.grab_json_response(','.join([str(x) for x in args]))
         props = props['items'][0]
         volumes = pool.volumes
         remainingcap = params['capacity']
@@ -551,14 +899,22 @@ class OEMHandler(generic.OEMHandler):
         parms = {'raidlink_AddNewVolWithNaAsync': arglist}
         rsp = self.wc.grab_json_response(url, parms)
         if rsp['return'] == 14:  # newer firmware
-            if cid[2] == 2:
-                cnum = cid[1]
-            arglist = '{0},{1},{2},{3},{4},{5},'.format(
-                cnum, params['raidlevel'], params['spans'],
-                params['perspan'], params['drives'], params['hotspares'])
-            arglist += ''.join(vols) + ',{0}'.format(cid[2])
-            parms = {'raidlink_AddNewVolWithNaAsync': arglist}
-            rsp = self.wc.grab_json_response(url, parms)
+            if 'supported_cpwb' in props and not usesctrlslot: # no ctrl_type
+                arglist = '{0},{1},{2},{3},{4},{5},{6},'.format(
+                    cnum, params['raidlevel'], params['spans'],
+                    params['perspan'], 0, params['drives'], params['hotspares'])
+                arglist += ''.join(vols)
+                parms = {'raidlink_AddNewVolWithNaAsync': arglist}
+                rsp = self.wc.grab_json_response(url, parms)
+            else: # with ctrl_type
+                if cid[2] == 2:
+                    cnum = cid[1]
+                arglist = '{0},{1},{2},{3},{4},{5},'.format(
+                    cnum, params['raidlevel'], params['spans'],
+                    params['perspan'], params['drives'], params['hotspares'])
+                arglist += ''.join(vols) + ',{0}'.format(cid[2])
+                parms = {'raidlink_AddNewVolWithNaAsync': arglist}
+                rsp = self.wc.grab_json_response(url, parms)
         if rsp['return'] != 0:
             raise Exception(
                 'Unexpected response to add volume command: ' + repr(rsp))
@@ -601,12 +957,30 @@ class OEMHandler(generic.OEMHandler):
             if pool.disks:
                 self._create_array(pool)
 
+    def weblogout(self):
+        if self._wc:
+            try:
+                self._wc.grab_json_response(self.logouturl)
+            except Exception:
+                pass
+            self._wc = None
+
     @property
     def wc(self):
-        if (not self._wc or (self._wc.vintage
-                             and self._wc.vintage < util._monotonic_time()
-                             - 30)):
-            self._wc = self.get_webclient()
+        while self.weblogging:
+            time.sleep(0.25)
+        self.weblogging = True
+        try:
+            if (not self._wc or (self._wc.vintage
+                                and self._wc.vintage < util._monotonic_time()
+                                - 30)):
+                if not self.updating and self._wc:
+                    # in case the existing session is still valid
+                    # dispose of the session
+                    self.weblogout()
+                self._wc = self.get_webclient()
+        finally:
+            self.weblogging = False
         return self._wc
 
     def get_webclient(self, login=True):
@@ -628,6 +1002,11 @@ class OEMHandler(generic.OEMHandler):
                    'Referer': referer,
                    'Host': 'xcc',
                    'Content-Type': 'application/json'}
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/providers/get_nonce', {})
+        if status == 200:
+            nonce = rsp.get('nonce', None)
+            headers['Content-Security-Policy'] = 'nonce={0}'.format(nonce)
         wc.request('POST', '/api/login', adata, headers)
         rsp = wc.getresponse()
         if rsp.status == 200:
@@ -640,6 +1019,11 @@ class OEMHandler(generic.OEMHandler):
                 wc.set_header('X-XSRF-TOKEN', wc.cookies['_csrf_token'])
             return wc
 
+    def grab_redfish_response_with_status(self, url, body=None, method=None):
+        wc = self.webclient.dupe()
+        res = wc.grab_json_response_with_status(url, body, method=method)
+        return res
+
     def list_media(self, fishclient):
         rt = self.wc.grab_json_response('/api/providers/rp_vm_remote_getdisk')
         if 'items' in rt:
@@ -673,18 +1057,31 @@ class OEMHandler(generic.OEMHandler):
                 'XCC does not have required license for operation')
 
     def upload_media(self, filename, progress=None, data=None):
-        xid = random.randint(0, 1000000000)
+        wc = self.wc
         self._refresh_token()
-        uploadthread = webclient.FileUploader(
-            self.wc, '/upload?X-Progress-ID={0}'.format(xid), filename, data)
+        rsp, statu = wc.grab_json_response_with_status('/rdocupload')
+        newmode = False
+        if statu == 404:
+            xid = random.randint(0, 1000000000)
+            uploadthread = webclient.FileUploader(
+                wc, '/upload?X-Progress-ID={0}'.format(xid), filename, data)
+        else:
+            newmode = True
+            uploadthread = webclient.FileUploader(
+                wc, '/rdocupload', filename, data)
         uploadthread.start()
         while uploadthread.isAlive():
             uploadthread.join(3)
-            rsp = self.wc.grab_json_response(
-                '/upload/progress?X-Progress-ID={0}'.format(xid))
-            if progress and rsp['state'] == 'uploading':
-                progress({'phase': 'upload',
-                          'progress': 100.0 * rsp['received'] / rsp['size']})
+            if newmode:
+                if progress:
+                    progress({'phase': 'upload',
+                          'progress': 100 * wc.get_upload_progress()})
+            else:
+                rsp = self.wc.grab_json_response(
+                    '/upload/progress?X-Progress-ID={0}'.format(xid))
+                if progress and rsp['state'] == 'uploading':
+                    progress({'phase': 'upload',
+                            'progress': 100.0 * rsp['received'] / rsp['size']})
             self._refresh_token()
         rsp = json.loads(uploadthread.rsp)
         if progress:
@@ -720,8 +1117,91 @@ class OEMHandler(generic.OEMHandler):
         if progress:
             progress({'phase': 'complete'})
 
+    def redfish_update_firmware(self, usd, filename, data, progress, bank):
+        if usd['HttpPushUriTargetsBusy']:
+            raise pygexc.TemporaryError('Cannot run multtiple updates to '
+                                        'same target concurrently')
+        upurl = usd['HttpPushUri']
+        self._do_web_request(
+            '/redfish/v1/UpdateService',
+            {'HttpPushUriTargetsBusy': True}, method='PATCH')
+        try:
+            if bank == 'backup':
+                self._do_web_request(
+                    '/redfish/v1/UpdateService',
+                    {'HttpPushUriTargets':
+                        ['/redfish/v1/UpdateService'
+                         '/FirmwareInventory/BMC-Backup']}, method='PATCH')
+            uploadthread = webclient.FileUploader(
+                self.webclient, upurl, filename, data, formwrap=False,
+                excepterror=False)
+            uploadthread.start()
+            wc = self.webclient
+            while uploadthread.isAlive():
+                uploadthread.join(3)
+                if progress:
+                    progress(
+                        {'phase': 'upload',
+                         'progress': 100 * wc.get_upload_progress()})
+            if (uploadthread.rspstatus >= 300
+                    or uploadthread.rspstatus < 200):
+                rsp = uploadthread.rsp
+                errmsg = ''
+                try:
+                    rsp = json.loads(rsp)
+                    errmsg = (
+                        rsp['error'][
+                            '@Message.ExtendedInfo'][0]['Message'])
+                except Exception:
+                    raise Exception(uploadthread.rsp)
+                raise Exception(errmsg)
+            rsp = json.loads(uploadthread.rsp)
+            monitorurl = rsp['@odata.id']
+            complete = False
+            phase = "apply"
+            statetype = 'TaskState'
+            while not complete:
+                pgress = self._do_web_request(monitorurl, cache=False)
+                if not pgress:
+                    break
+                for msg in pgress.get('Messages', []):
+                    if 'Verify failed' in msg.get('Message', ''):
+                        raise Exception(msg['Message'])
+                state = pgress[statetype]
+                if state in ('Cancelled', 'Exception', 'Interrupted',
+                             'Suspended'):
+                    raise Exception(
+                        json.dumps(json.dumps(pgress['Messages'])))
+                pct = float(pgress['PercentComplete'])
+                complete = state == 'Completed'
+                progress({'phase': phase, 'progress': pct})
+                if complete:
+                    if 'OperationTransitionedToJob' in pgress['Messages'][0]['MessageId']:
+                        monitorurl = pgress['Messages'][0]['MessageArgs'][0]
+                        phase = 'validating'
+                        statetype = 'JobState'
+                        complete = False
+                        time.sleep(3)
+                else:
+                    time.sleep(3)
+            if bank == 'backup':
+                return 'complete'
+            return 'pending'
+        finally:
+            self._do_web_request(
+                '/redfish/v1/UpdateService',
+                {'HttpPushUriTargetsBusy': False}, method='PATCH')
+            self._do_web_request(
+                '/redfish/v1/UpdateService',
+                {'HttpPushUriTargets': []}, method='PATCH')
+
     def update_firmware(self, filename, data=None, progress=None, bank=None):
         result = None
+        usd = self._do_web_request('/redfish/v1/UpdateService')
+        rfishurl = usd.get('HttpPushUri', None)
+        if rfishurl:
+            return self.redfish_update_firmware(
+                usd, filename, data, progress, bank)
         if self.updating:
             raise pygexc.TemporaryError('Cannot run multiple updates to same '
                                         'target concurrently')
@@ -1016,6 +1496,17 @@ class OEMHandler(generic.OEMHandler):
                     license_errors[rsp['return']])
         return self.get_licenses()
 
+    def user_delete(self, uid):
+        userinfo = self.wc.grab_json_response('/api/dataset/imm_users')
+        uidtonamemap = {}
+        for user in userinfo.get('items', [{'users': []}])[0].get('users', []):
+            uidtonamemap[user['users_user_id']] = user['users_user_name']
+        if uid in uidtonamemap:
+            deltarget = '{0},{1}'.format(uid, uidtonamemap[uid])
+            self.wc.grab_json_response('/api/function', {"USER_UserDelete": deltarget})
+            return True
+        return super(OEMHandler, self).user_delete(uid)
+
     def get_user_expiration(self, uid):
         userinfo = self.wc.grab_json_response('/api/dataset/imm_users')
         for user in userinfo['items'][0]['users']:
